!#################################################################
!
! MODULE: emission_co2_ff 
! 
! PUBLIC TYPES:
!
! PUBLIC MEMBER FUNCTIONS:
!
!   emission_co2_ff_init
!   emission_co2_ff_done
!   emission_co2_ff_calc
!
! PUBLIC DATA MEMBERS:
!
! DESCRIPTION: 
!
!   Compute CO2 fossil fuel emissions in kgC/gridbox/mo.
!
! REVISION HISTORY: 
!
!   Mike Trudeau, Nov 2011
!   Read monthly global 1x1 ODIAC CO2 fossil fuel emissions. 
!
!### macros #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
!#################################################################

module emission_co2_ff
  
  use GO,                  only : gol, goErr, goPr
  use emission_common,     only : enkf_parms, field_1x1, EmisInputDir
  use emission_common,     only : nmembersloc
  use Dims,                only : nlon360, nlat180

  implicit none

  character(len=*), parameter   :: mname = 'emission_co2_ff'

  private

  character(len=225)           :: ff_prefix
  character(len=256)            :: ff_varname

  logical					:: ff_TIMES_weekly
  character(len=1024)				:: ff_TIMES_weekly_ncfile
  real*4, dimension(:,:,:), allocatable		:: ff_TIMES_weekly_factors

  logical					:: ff_TIMES_diurnal
  character(len=1024)           		:: ff_TIMES_diurnal_ncfile
  real*4, dimension(:,:,:), allocatable		:: ff_TIMES_diurnal_factors

  logical					:: ff_TIMES_verbose

  public                        :: emission_co2_ff_init, emission_co2_ff_done, emission_co2_ff_calc

  character(len=2048)           :: FFInputDir ! default is EmisInputDir

contains

  subroutine emission_co2_ff_init( status )

    use GO,			only : TrcFile, Init, Done, ReadRc
    use global_data,		only : rcfile
    use MDF, 			only : MDF_NETCDF
    use ReadFromFile_MDF,	only : ReadFromFile

    implicit none

    character(len=*), parameter :: rname = mname//'/emission_co2_ff_init'

    integer, intent(out)        :: status
    type(TrcFile)               :: rcF

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'ff.input.dir', FFInputDir, status, default = EmisInputDir)
    IF_ERROR_RETURN(status=1)
    call ReadRc( rcF, 'ff.ncfile.prefix', ff_prefix, status, ff_prefix)
    IF_ERROR_RETURN(status=1)
    call ReadRc( rcF, 'ff.ncfile.varname', ff_varname , status, 'fossil')
    IF_ERROR_RETURN(status=1)

    call ReadRc( rcF, 'ff.TIMES.verbose', ff_TIMES_verbose , status, .true.)
    IF_ERROR_RETURN(status=1)

    call ReadRc( rcF, 'ff.TIMES.weekly', ff_TIMES_weekly , status, .false.)
    IF_ERROR_RETURN(status=1)
    
    if(ff_TIMES_weekly) then

       call ReadRc( rcF, 'ff.TIMES.weekly.ncfile', ff_TIMES_weekly_ncfile, status, &
            "/home/Andy.Jacobson/Data/emissions/Nassar-TIMES/weekly_scale_factors_constant_China.1x1.nc")
       IF_ERROR_RETURN(status=1)

       allocate(ff_TIMES_weekly_factors(nlon360,nlat180,7))

      call ReadFromFile(fname = ff_TIMES_weekly_ncfile, & 
                        ftype = MDF_NETCDF, vname = 'weekly_scale_factors', &
                        buffer = ff_TIMES_weekly_factors, status = status)
      IF_NOTOK_RETURN(status=1)
       
    endif

    call ReadRc( rcF, 'ff.TIMES.diurnal', ff_TIMES_diurnal , status, .false.)
    IF_ERROR_RETURN(status=1)

    if(ff_TIMES_diurnal) then

       call ReadRc( rcF, 'ff.TIMES.diurnal.ncfile', ff_TIMES_diurnal_ncfile, status, &
            "/home/Andy.Jacobson/Data/emissions/Nassar-TIMES/diurnal_scale_factors.1x1.nc")
       IF_ERROR_RETURN(status=1)

       allocate(ff_TIMES_diurnal_factors(nlon360,nlat180,8)) ! 3-hourly is 8 slots

       call ReadFromFile(fname = ff_TIMES_diurnal_ncfile, & 
            ftype = MDF_NETCDF, vname = 'diurnal_scale_factors', &
            buffer = ff_TIMES_diurnal_factors, status = status)
       IF_NOTOK_RETURN(status=1)
       
    endif

    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)
    
    write (gol,'("[emission_co2_ff_init]	FF input files to be read from """,a,""",")') trim(FFInputDir)
    call goPr
    write (gol,'("[emission_co2_ff_init]	   with filename prefix: """,a,""",")') trim(ff_prefix)
    call goPr
    write (gol,'("[emission_co2_ff_init]	   and variable name:  """,a,""".")') trim(ff_varname)
    call goPr
    if(ff_TIMES_diurnal) then
       write (gol,'("[emission_co2_ff_init]	   diurnal scale factors from:  """,a,""".")') trim(ff_TIMES_diurnal_ncfile)
       call goPr
    endif
    if(ff_TIMES_weekly) then
       write (gol,'("[emission_co2_ff_init]	   weekly scale factors from:  """,a,""".")') trim(ff_TIMES_weekly_ncfile)
       call goPr
    endif
    status = 0

  end subroutine emission_co2_ff_init

  subroutine emission_co2_ff_calc( parms, co2_ff, flux_updated, parms_updated, status )

    use dims,                              only : idate, newmonth, newsrun, nlon360, nlat180, sec_month, dxy11, sec_year
    use MDF,                               only : MDF_NETCDF
    use ReadFromFile_MDF,                  only : ReadFromFile
    use go_string,                         only : goNum2str
    use go_date,			   only : calc_days_in_month, calc_weekday

    implicit none
    
    character(len=*), parameter                 :: rname = mname//'/emission_co2_ff_calc'
    type(enkf_parms), dimension(:), intent(in)  :: parms ! not currently used as fossil fuel emissions are not optimized
    real*4, dimension(nlon360, nlat180, 12)       :: ncArr
    integer                                     :: i, j, t ! lon, lat, time
    type(field_1x1), dimension(:), intent(out)  :: co2_ff
    integer, intent(out)                        :: status
    logical, intent(out)                        :: flux_updated
    logical, intent(in)	                        :: parms_updated
    integer                                     :: imember
    integer                                     :: greenwich_weekday,this_weekday,localday,islot
    real					:: greenwich_hours,hrperdeg,lon
    real					:: month2year
    real					:: spm ! seconds per month
    real,dimension(nlon360,nlat180)		:: TIMES_scaling_factors
    character(len=3), dimension(7) 		:: daynames  = (/'Sun','Mon','Tue','Wed','Thu','Fri','Sat'/)

    ! array ff_input (360,180,12) is persistent.

    flux_updated = .false.

    if ( newmonth .or. newsrun ) then
 
      flux_updated = .true.

      call ReadFromFile(fname = trim(FFInputDir)//'/'//trim(ff_prefix)//'_'//trim(goNum2str(idate(1), '(i4.4)'))//'.nc', & 
                        ftype = MDF_NETCDF, vname = trim(ff_varname), &
                        buffer = ncArr, status = status)
      IF_NOTOK_RETURN(status=1)

      do j = 1, nlat180
        ncArr(:, j, idate(2)) = ncArr(:, j, idate(2)) * sec_month / 86400.0 ! gC/m2/day to gC/m2/mo
        ncArr(:, j, idate(2)) = ncArr(:, j, idate(2)) * 1.0E-3 ! gC/m2/mo to kgC/m2/mo
        ncArr(:, j, idate(2)) = ncArr(:, j, idate(2)) * dxy11(j) ! kgC/m2/mo to kgC/gridbox/mo
      enddo

      if(ff_TIMES_verbose) then
         write (gol,'("[emission_co2_ff_calc]	At idate ",i4,x,5(i2.2,x),", NEWMONTH")') idate
         call goPr
      endif

   endif

   TIMES_scaling_factors=1.

   if (ff_TIMES_weekly) then
      greenwich_weekday = calc_weekday(idate(1),idate(2),idate(3))
      greenwich_hours=idate(4)+idate(5)/60.
      hrperdeg=24/360.
      ! N.B. the TIMES_weekly data use the following indexing:
      !  1 = Mon, 2 = Tue, ... , 7 = Sun
      ! The weekday function returns 1 = Sun, 2 = Mon, ..., 7 = Sat
      
      if(ff_TIMES_verbose) then
         write (gol,'("[emission_co2_ff_calc]	At idate ",i4,x,5(i2.2,x),", weekday is ",a," (",i1,") with hours ",f7.2)') &
              idate,daynames(greenwich_weekday),greenwich_weekday,greenwich_hours
         call goPr
      endif
      
      do i=1,nlon360
         lon=-179.5+(i-1)*1.0
         this_weekday=greenwich_weekday
         if (greenwich_hours + hrperdeg*lon < 0) then
            this_weekday = greenwich_weekday - 1
            if (this_weekday < 1 ) then
               this_weekday = 7
            endif
            if(ff_TIMES_verbose) then
               write (gol,'("[emission_co2_ff_calc]	At lon ",f6.1,", weekday is ",a," (",i1,") with hours ",f7.2)') &
                    lon,daynames(this_weekday),this_weekday,greenwich_hours+hrperdeg*lon
               call goPr
            endif
         endif
         if (greenwich_hours + hrperdeg*lon >= 24) then
            this_weekday = greenwich_weekday + 1
            if (this_weekday > 7 ) then
               this_weekday = 1
            endif
            if(ff_TIMES_verbose) then
               write (gol,'("[emission_co2_ff_calc]	At lon ",f6.1,", weekday is ",a," (",i1,") with hours ",f7.2)') &
                    lon,daynames(this_weekday),this_weekday,greenwich_hours+hrperdeg*lon
               call goPr
            endif
         endif
         localday=this_weekday - 1 ! convert to TIMES indexing of days
         if (localday < 1 ) then
            localday = 7
         endif
         TIMES_scaling_factors(i,:)=TIMES_scaling_factors(i,:) * ff_TIMES_weekly_factors(i,:,localday)
      enddo
      flux_updated = .true.
   endif

   if (ff_TIMES_diurnal) then
      
      islot=1+idate(4)/3 ! integer truncating arithmetic 1+0:23/3 -> 1 1 1 2 2 2 .... 8 8 8 
      if(ff_TIMES_verbose) then
         write (gol,'("[emission_co2_ff_calc]	TIMES_diurnal: hour ",i2," is 3-hr slot ",i2,".")') idate(4),islot
         call goPr
      endif
      TIMES_scaling_factors=TIMES_scaling_factors * ff_TIMES_diurnal_factors(:,:,islot) ! 
      flux_updated = .true.
   endif

   ! ensemble members assigned same fossil fuel flux
   month2year = sec_year/sec_month
   do imember=1,nmembersloc
      co2_ff(imember)%surf = ncArr(:, :, idate(2))*TIMES_scaling_factors
      
      write (gol,'("[emission_co2_ff_calc]	member ",i4,": fossil fuels emission global integral ",f12.4," PgC/yr.")') imember, &
           sum(co2_ff(imember)%surf * 1.0E-12 *month2year)
      call goPr
   end do

   status = 0

  end subroutine emission_co2_ff_calc

  subroutine emission_co2_ff_done(status)

    implicit none

    integer, intent(out)        :: status
    character(len=*), parameter :: rname = mname//'/emission_co2_ff_done'

    status=0

  end subroutine emission_co2_ff_done

end module emission_co2_ff
