!### macros #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
! Ported from similar code in CT emissions_common.F90 by Andy Jacobson, 28 May 2014
!
!
module user_output_flux1x1

  use GO,				only : gol, goErr, goPr, goLabel
  use Dims,            	      		only : nlon360,nlat180
  use chem_param, 			only : ntrace_emis

  ! CT co2-specific
  use emission_common,			only : co2_bio,co2_ocn,co2_ff,co2_fires,field_1x1

  implicit none

  !  This module expects a structure of the following sort, to hold
  !  the accumulated, weighted field (surf), the accumulated weights
  !  (wda), and the number of samples (n).  For CT-CO2, we use a
  !  structure provided by module emission_common.  Other codes may
  !  need to define this structure.
  !
  !  type field_1x1
  !     real*8, dimension(nlon360,nlat180)			:: surf
  !     integer						:: n
  !     real*8						:: wda
  !  end type field_1x1


  ! This f1x1_container structure is CT-CO2 specific.  This is a means
  ! of keeping an array of pointers.  Apparently they must be inside a
  ! container structure.
  type f1x1_container
     type(field_1x1), pointer :: f1x1
  end type f1x1_container
  type(f1x1_container), dimension(:),allocatable	:: raw_emis ! CT-specific access to raw 1x1 flux data


  character*20, dimension(:), allocatable		:: flux_names ! variable names for netCDF file
  type(field_1x1), dimension(:), allocatable		:: emis_accum ! accumulator structure, per flux

  integer, dimension(6)					:: idate_api=-1 ! averaging period start
  integer, dimension(6)					:: idate_ape=-1 ! averaging period end

  integer					        :: dhour         ! every dhour hour

  private

  public						:: idate_api,idate_ape
  public						:: user_output_flux1x1_init
  public						:: user_output_flux1x1_done
  public						:: user_output_flux1x1_accumulate
  public						:: user_output_flux1x1_calc
  public						:: user_output_flux1x1_reset
  public						:: user_output_flux1x1_write

  character(len=*), parameter  ::  mname = 'user_output_flux1x1'

contains

  subroutine user_output_flux1x1_init(rcF,status)

    use GO,                     only : TrcFile, Init, Done, ReadRc

    implicit none

    character(len=*), parameter		:: rname = mname//'/user_output_flux1x1_init'
    integer,intent(out)			:: status
    type(TrcFile)	                :: rcF

    ! user_output also reads and uses the dhour variable.  We actually
    ! only need it for documentary purposes (it is output in the
    ! netCDF file).  However, it is not possible to "use" it from
    ! user_output without causing circular references.  Thus we
    ! re-read it here.  We could also just pass it to the "write"
    ! routine from user_output.

    call ReadRc( rcF, 'output.flux1x1.dhour', dhour, status , default=24)
    IF_NOTOK_RETURN(status=1)

    allocate(emis_accum(ntrace_emis))
    allocate(flux_names(ntrace_emis))

    allocate(raw_emis(ntrace_emis)) ! CT-specific

    ! CT-specific variable names
    flux_names(1)='bio_flux_opt'
    flux_names(2)='ocn_flux_opt'
    flux_names(3)='fossil_flux_imp'
    flux_names(4)='fire_flux_imp'

    ! CT-specific access to raw 1x1 flux data
    raw_emis(1)%f1x1 => co2_bio(1)
    raw_emis(2)%f1x1 => co2_ocn(1)
    raw_emis(3)%f1x1 => co2_ff(1)
    raw_emis(4)%f1x1 => co2_fires(1)

    ! make sure we start out with correct accumulator status
    call user_output_flux1x1_reset(status)
    IF_NOTOK_RETURN(status=1)

    status=0

  end subroutine user_output_flux1x1_init



  subroutine user_output_flux1x1_reset(status)

    ! Reset the accumulators to initial state.

    implicit none

    character(len=*), parameter		:: rname = mname//'/user_output_flux1x1_reset'
    integer,intent(out)			:: status
    integer				:: itrace

    do itrace = 1,ntrace_emis
       emis_accum(itrace)%surf = 0.
       emis_accum(itrace)%n = 0
       emis_accum(itrace)%wda = 0.
    end do

    status=0

  end subroutine user_output_flux1x1_reset



  subroutine user_output_flux1x1_done(status)

    ! De-allocate arrays and perform any other needed tidying up in
    ! preparation for termination.

    implicit none

    character(len=*), parameter		:: rname = mname//'/user_output_flux1x1_done'
    integer,intent(out)			:: status

    ! user_output should have called user_output_flux1x1_write for the
    ! last period.

    deallocate(emis_accum)
    deallocate(flux_names)

    status=0

  end subroutine user_output_flux1x1_done


  subroutine user_output_flux1x1_accumulate(region,status)

    ! Sample 1x1 fluxes and accumulate ndyn-weighted averages.

    use dims,			only : itaur, itaui, ndyn, ndyn_max,idate
    use dims,			only : sec_month, dxy11, nlat180
    use chem_param,		only : xmc

    implicit none

    integer,intent(in)			:: region
    character(len=*), parameter		:: rname = mname//'/user_output_flux1x1_accumulate'
    integer,intent(out)			:: status
    integer				:: itrace
    real				:: weight, convfac
    integer				:: j

    ! If not global region, do nothing.  We manage emissions on a
    ! global 1x1 grid only, so we only need to do this once for all
    ! regions.

    if(region .eq. 1) then

       weight = real(ndyn)/real(ndyn_max)

       ! accumulate average fluxes (wrap in if myid .eq. root for ensemble)
       ! at this time, change from kgC cell-1 month-1 to mol m-2 s-1

       do j=1,nlat180

          convfac = weight*1.e3/(xmc*dxy11(j)*real(sec_month))  ! CO2 specific

          do itrace=1,ntrace_emis
             ! note CT-specific access to 1x1 data on next line
             emis_accum(itrace)%surf(:,j) = emis_accum(itrace)%surf(:,j) + raw_emis(itrace)%f1x1%surf(:,j)*convfac

          end do
 
       end do

       do itrace=1,ntrace_emis

          emis_accum(itrace)%n = emis_accum(itrace)%n + 1
          emis_accum(itrace)%wda = emis_accum(itrace)%wda + weight

       end do

    end if

    status=0

  end subroutine user_output_flux1x1_accumulate

  subroutine user_output_flux1x1_calc(status)

    ! Compute average; note that units are already mol m-2 s-1 
    ! 
    ! Note that this is safe against multiple calls because it sets n
    ! to zero when it performs the evaluation, and we don't do
    ! anything if n is zero.

    integer,intent(out)			:: status
    integer				:: itrace

    ! compute average; note that units are already mol m-2 s-1 

    do itrace=1,ntrace_emis

       if(emis_accum(itrace)%n .gt. 0) then
          emis_accum(itrace)%surf = emis_accum(itrace)%surf / emis_accum(itrace)%wda
          emis_accum(itrace)%n = 0
          emis_accum(itrace)%wda = 0.
       end if

    end do

    status=0

  end subroutine user_output_flux1x1_calc

  subroutine user_output_flux1x1_write(status)

    ! Write netCDF output file.  Calls evaluate first, then writes,
    ! then calls reset.

    use ParTools, 	only : myid, root
    use MDF, 		only : MDF_Open
    use MDF, 		only : MDF_Create, MDF_Close, MDF_EndDef
    use MDF, 		only : MDF_NETCDF, MDF_REPLACE, MDF_GLOBAL, MDF_CHAR, MDF_INT, MDF_FLOAT, MDF_UNLIMITED, MDF_READ, MDF_DOUBLE
    use MDF, 		only : MDF_Put_Att
    use MDF, 		only : MDF_Def_Dim
    use MDF, 		only : MDF_Inq_Dimid, MDF_Inquire_Dimension
    use MDF, 		only : MDF_Inq_VarID
    use MDF, 		only : MDF_Def_Var, MDF_Put_Var, MDF_Get_Var
    use global_data, 	only : outdir
    use dims, 		only : itau,itaui
    use datetime, 	only : tau2date, date2tau, idate2ddate

    implicit none

    integer,intent(out)			:: status
    integer				:: itrace
    character(len=200)			:: attname,comp,pp
    character(len=2048)			:: FluxOutFile = ''

    character(len=*), parameter		:: rname = mname//'/user_output_flux1x1_write'

    integer				:: i
    integer				:: hid
    integer				:: dim_lon_id, dim_lat_id, dim_date_id, dim_date_comps
    integer,dimension(:),allocatable	:: var_flux
    integer				:: var_lon, var_lat, var_date, var_comps, var_decdate
    integer, dimension(6)		:: idate_epoch, idate_middle
    integer				:: itau_epoch, itau_api, itau_ape
    real				:: itau_middle ! real since it is an average
    real				:: date_value ! days-since value corresponding to itau_middle, to be written
    real*8				:: decimal_date
    character(len=256)			:: progstring
    character(len=1024)			:: attstring
    real,dimension(2)			:: axis_range
    real*8,parameter			:: fillval_r8 = -1.e34
    integer, dimension(8)		:: isysdate
    character(len=256)			:: sysdate 

    if(myid .ne. root) return

    ! Check to see if anything has been accumulated since last
    ! write/reset.  If not, return.

    if(emis_accum(1)%n .eq. 0) then
       write (gol,'("[user_output_flux1x1_write]  No samples accumulated--no output file created.")'); call goPr
       return
    end if

    allocate(var_flux(ntrace_emis))

    if(idate_api(1) .eq. -1) then
       call tau2date(itaui,idate_api) ! initialize
    endif

    call user_output_flux1x1_calc(status) ! OK if called repeatedly
    IF_NOTOK_RETURN(status=1)

    call tau2date(itau,idate_ape) ! set averaging period end to right now
    itau_ape=itau
    call date2tau(idate_api,itau_api)

    idate_epoch = (/2000,1,1,0,0,0/)
    call date2tau(idate_epoch,itau_epoch)

    itau_middle = 0.5*(itau_api + itau_ape) - itau_epoch
    date_value = itau_middle/86400.

    itau_middle = 0.5*(itau_api + itau_ape) ! note recycling of itau_middle here
    call tau2date(int(itau_middle),idate_middle)

    decimal_date=idate2ddate(idate_middle)

    write (FluxOutFile,'(a,"/flux1x1_",i4.4,4i2.2,"_",i4.4,4i2.2,".nc")') &
                trim(outdir), idate_api(1:5), idate_ape(1:5)

    call MDF_Create( trim(FluxOutFile), MDF_NETCDF, MDF_REPLACE, hid, status )
    IF_NOTOK_RETURN(status=1)

    ! Global Attributes 

    write(attstring,'(i4.4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2, " UTC")') idate_api
    call MDF_Put_Att(hid, MDF_GLOBAL,"averaging_period_start_date",values=trim(attstring),status=status)
    IF_NOTOK_RETURN(status=1)

    write(attstring,'(i4.4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2, " UTC")') idate_ape
    call MDF_Put_Att(hid, MDF_GLOBAL,"averaging_period_end_date",values=trim(attstring),status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Att(hid, MDF_GLOBAL,"averaging_period_length_hours",values=dhour,status=status)
    IF_NOTOK_RETURN(status=1)


    call MDF_Put_Att(hid, MDF_GLOBAL,"email",values="carbontracker.team@noaa.gov",status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Att(hid, MDF_GLOBAL,"url",values="http://carbontracker.noaa.gov",status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Att(hid, MDF_GLOBAL,"institution",values="NOAA Earth System Research Laboratory",status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Att(hid, MDF_GLOBAL,"Conventions",values="CF-1.5",status=status)
    IF_NOTOK_RETURN(status=1)

    call getarg (0, progstring, status)
    call date_and_time(values = isysdate)
    write (sysdate, '(i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," UTC")') &
           isysdate(1), isysdate(2), isysdate(3), isysdate(5), isysdate(6), isysdate(7)
    write(attstring,'("Created ",a," by ",a,".")') trim(sysdate),trim(progstring)
    call MDF_Put_Att(hid, MDF_GLOBAL,"history",values=attstring,status=status)
    IF_NOTOK_RETURN(status=1)

    ! define dimensions:                                                                                                                                                                                            
    call MDF_Def_Dim( hid, 'lon', 360 , dim_lon_id, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Def_Dim( hid, 'lat', 180 , dim_lat_id, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Def_Dim( hid, 'date', MDF_UNLIMITED , dim_date_id, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Def_Dim( hid, 'date_components', 6 , dim_date_comps, status )
    IF_NOTOK_RETURN(status=1)

    ! dimension variables:                                                                                                                                                                                                    
    call MDF_Def_Var( hid, 'lon', MDF_DOUBLE, (/dim_lon_id/), var_lon, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_lon,"units",values="degrees_east",status=status)
    IF_NOTOK_RETURN(status=1)
    axis_range = (/ -180., 180./)
    call MDF_Put_Att(hid, var_lon,"actual_range",values=axis_range,status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var( hid, 'lat', MDF_DOUBLE, (/dim_lat_id/), var_lat, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_lat,"units",values="degrees_north",status=status)
    IF_NOTOK_RETURN(status=1)
    axis_range = (/ -90., 90./)
    call MDF_Put_Att(hid, var_lat,"actual_range",values=axis_range,status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var( hid, 'date', MDF_DOUBLE, (/dim_date_id/), var_date, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_date,"units",values="days since 2000-01-01 00:00:00 UTC",status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var( hid, 'decimal_date', MDF_DOUBLE, (/dim_date_id/), var_decdate, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_decdate,"units",values="years",status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var( hid, 'date_components', MDF_INT, (/dim_date_comps, dim_date_id/), var_comps, status )
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_comps,"units",values="NA",status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_comps,"comment",values="Calendar date components as integers.  Times and dates are UTC.",status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(hid, var_comps,"order",values="year, month, day, hour, minute, second",status=status)
    IF_NOTOK_RETURN(status=1)

    ! define variables:                                                                                                                                                                  
    do itrace=1,ntrace_emis
       call MDF_Def_Var( hid, trim(flux_names(itrace)), MDF_DOUBLE, (/dim_lon_id,dim_lat_id,dim_date_id/), var_flux(itrace), deflate_level=9, status=status )
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(hid, var_flux(itrace),"units",values="mol m-2 s-1",status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(hid, var_flux(itrace),"_FillValue",values=fillval_r8,status=status)
       IF_NOTOK_RETURN(status=1)

    end do

    ! finished definition:                                                                                                                                                                                          
    call MDF_EndDef( hid, status )
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var( hid, var_lon, (/(i+0.5,i=-180,179)/), status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Var( hid, var_lat, (/(i+0.5,i=-90,89)/), status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Var( hid, var_date, (/(date_value)/), status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Var( hid, var_decdate, (/(decimal_date)/), status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var( hid, var_comps, (/(idate_middle)/), status)
    IF_NOTOK_RETURN(status=1)

    do itrace=1,ntrace_emis
       call MDF_Put_Var( hid, var_flux(itrace), emis_accum(itrace)%surf, status)
       IF_NOTOK_RETURN(status=1)
    end do

    ! close file:                                                                                                                                                                                                   
    call MDF_Close( hid, status )
    IF_NOTOK_RETURN(status=1)

    call tau2date(itau,idate_api) ! for next averaging period

    call user_output_flux1x1_reset(status)
    IF_NOTOK_RETURN(status=1)

    write (gol,'("[user_output_flux1x1_write]  Done writing arrays and closing output file.")'); call goPr

    deallocate(var_flux)

    status=0

  end subroutine user_output_flux1x1_write

end module user_output_flux1x1
