!#################################################################
!
! MODULE: user_output_sounding
!
! PUBLIC SUBROUTINES: 
!   user_output_sounding_init
!   user_output_sounding_step
!   user_output_sounding_done
!
! PRIVATE SUBROUTINES: 
!   user_output_sounding_sample
!   user_output_sounding_write
!   user_output_sounding_read
!   user_output_sounding_reset
!
! DESCRIPTION: 
!   Write column tracer fields on interpolated levels at discrete
!   sample times. Called from user_output.F90. 
!
! REVISION HISTORY: 
!
!   Andy Jacobson, Oct 2014 - added specific humidity to output
!
!   Mike Trudeau, Jan 2014
!     Created to read GOSAT-ACOS formatted NetCDF files ("acos_b34_L2lite_r02c")
!     and write NOAA CarbonTracker formatted NetCDF output.

!   Andy Jacobson, Oct 2014 - added specific humidity to output

!   Mike Trudeau, Feb 2015
!     Modified to read GOSAT-ACOS formatted NetCDF files ("acos_b35_L2lite_r02")
!     and write NOAA CarbonTracker formatted NetCDF output. Removed "bg"
!     variable from output. All other CO2 tracer variables re-written to
!     contain a background component and a tracer component to the mole
!     fraction.
!
!   Andy Jacobson, Apr 2015 - merged Mike's b3.5-format code with 
!     my branch to preserve humidity output and other tweaks.
!
!   Andy Jacobson, Oct 2015 - modified to work with OCO-2 Lite format files,
!     including new use of int64 output for the sounding_id from mdf module.
!
!   Andy Jacobson, Dec 2015 - removed reads of vertical info from input files
!     reworked input to use ReadFromFile (allows int -> int64 promotion
!     made to work with parallel apps
!
!     Notes:
!     1. Developed with NF90 library to read input files. MDF
!        not currently capable of handling netCDF4 groups or int64.
!     2. "user_output_sounding_step" samples after the first call to vertical
!        diffusion "v" in the first half of the operator splitting sequence.
!        All soundings with a timestamp within the entire window of both
!        splitting sequences are sampled. The sampling is done only after the 
!        first call to "v" because it follows the "xyz" advection sequence, with 
!        the idea that whatever chemistry has been applied will have seem some 
!        influence of transport. Calls to user_output_sounding_step after 
!        source/sink/chemistry (SCS) sequences are ignored. Note that sampling 
!        after vertical diffusion is defined as an rc-key ("output.after.step : v").
!
!        ----------
!        Operator sequence, Table 1 (Krol et al., 2005).
!        * denotes a sample of the model
!        ----------
!                 t               t+dT/2
!        region 1 XYZ             V*S
!        region 2    XYZV*SCSVZYX
!                 t+dT/2                t+dT
!        region 1 CSVZYX
!        region 2         CSVZYXXYZVS
!
!     Required rc file keys:
!       output.sounding : [T/F]
!       output.sounding.indir   : [/path/to/netcdf/files]
!       output.sounding.ncfile   : [prefix]"<YYYYMMDD>"[suffix]".nc"
!
!     Example:
!       output.sounding : T
!       output.sounding.indir : /scratch2/portfolios/BMC/co2/input/acos
!       output.sounding.ncfile : acos_b35_L2lite_<YYYYMMDD>_r02.nc
!
!### macros #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
#define IF_NF90_NOTOK_RETURN(action) if (status/=NF90_NOERR) then; TRACEBACK; action; return; end if
#define IF_NOTOK_MDF(action) if (status/=0) then; TRACEBACK; action; call MDF_CLose(fid,status); status=1; return; end if 

!
#include "tm5.inc"
!
!#################################################################

module user_output_sounding

  use dims,        only : nregions
  use GO,          only : gol, goErr, goPr
  use ParTools,    only : myid, root, npes, par_barrier, ntracetloc, ntracet_ar
  use chem_param,  only : names, ntracet
  use NetCDF,      only : NF90_DOUBLE, NF90_INT64
  
  implicit none

  private

  ! --- public routines ------------------------

  public :: user_output_sounding_init
  public :: user_output_sounding_step
  public :: user_output_sounding_done

  ! --- constants ------------------------------

  character(len=*), parameter ::  mname = 'user_output_sounding'
  logical         , parameter :: debug = .false.
  integer         , parameter :: interp_method = 2 ! interpolation method output to file
                                          !   1: no interpolation ('grd')
                                          !   2: interpolation of tracers based on slopes ('slp')
                                          !   3: 3D linear interpolations ('int')
  
  ! --- variables ------------------------------

  character(len=1024) :: indir
  character(len=1024) :: fname 
  integer             :: nsoundings
  integer             :: sid_xtype ! NF90 xtype for sounding_id (NF90_DOUBLE, NF90_INT64)
  integer             :: fid ! file id for IF_NOTOK_MDF macro   

  type soundings_fields

     ! In some sounding input files, sounding_id is an int64, in
     ! others it is r8.  Accommodate both types, taking care to
     ! store/write out same type as we read.

    integer*8, dimension(:), allocatable :: sounding_id_i8
    real*8   , dimension(:), allocatable :: sounding_id_r8

    integer          , dimension(:)    , allocatable :: itau_sample 
    integer          , dimension(6)                  :: date 
    integer                                          :: tau
    real(8)          , dimension(:)    , allocatable :: sec_1970 
    integer          , dimension(:,:)  , allocatable :: idate
    character(len=10), dimension(:)    , allocatable :: region
    real             , dimension(:)    , allocatable :: latitude 
    real             , dimension(:)    , allocatable :: longitude
    integer          , dimension(:)    , allocatable :: sounding_num
    logical          , dimension(:)    , allocatable :: mask
    real             , dimension(:)    , allocatable :: oro ! orography
    real             , dimension(:)    , allocatable :: blh ! boundary layer height (m)
    real             , dimension(:,:)  , allocatable :: temper
    real             , dimension(:,:)  , allocatable :: phlb 
    real             , dimension(:,:,:), allocatable :: mf
    real             , dimension(:,:)  , allocatable :: m
    real             , dimension(:,:)  , allocatable :: q
    real             , dimension(:,:)  , allocatable :: gph ! geopotential height
    real             , dimension(:,:)  , allocatable :: wu ! zonal component, wind velocity (x-direction)
    real             , dimension(:,:)  , allocatable :: wv ! meridional component, wind velocity (y-direction)
  endtype soundings_fields

  type(soundings_fields) :: soundings 

  ! --- begin ----------------------------------

  contains

  subroutine user_output_sounding_init(rcF, status)

    ! --------------------------------------------
    ! Read the rc file
    ! --------------------------------------------

    use GO,     only : TrcFile, ReadRc
    use dims,   only : lm

    implicit none

    ! --- variables ------------------------------

    type(TrcFile), intent(in)  :: rcF
    integer      , intent(out) :: status
 
    ! --- constants ------------------------------
    
    character(len=*), parameter :: rname = mname//'/user_output_sounding_init'
    
    ! --- begin ----------------------------------

    call ReadRc(rcF, 'output.sounding.indir', indir, status)
    IF_NOTOK_RETURN(status=1)

    call ReadRc(rcF, 'output.sounding.ncfile', fname, status)
    IF_NOTOK_RETURN(status=1)

    call user_output_sounding_read(status)
    IF_NOTOK_RETURN(status=1)

    ! ok
    status = 0

  end subroutine user_output_sounding_init

  subroutine user_output_sounding_read(status)

    ! --------------------------------------------
    ! Read the input file for latitude, longitude, altitude (pressure), and time. 
    ! --------------------------------------------

    use datetime,           only : date2tau, tau2date
    use dims,               only : idate, itau, lm
    use NetCDF,             only : NF90_Open, NF90_Close, NF90_Inq_DimID, NF90_Inquire_Dimension
    use NetCDF,             only : NF90_Inq_NCID, NF90_Inq_VarID, NF90_Get_Var 
    use NetCDF,             only : NF90_NOWRITE, NF90_NOERR, NF90_Inquire_Variable
    use ReadFromFile_MDF

    implicit none

    ! --- variables ------------------------------

    integer, intent(out)  :: status
    character(len=1024)   :: infile = ''
    logical               :: exists_infile
    integer               :: ncid, dimid, varid
    integer               :: isound
    integer               :: itau_1970
    real*8                :: sec_sample
    integer, dimension(6) :: date_sample
  
    ! --- constants ------------------------------
    
    integer         , parameter :: region = 1 ! reference region
    character(len=*), parameter :: rname = mname//'/user_output_sounding_read'
    
    ! --- begin ----------------------------------

    write (infile,'(a,"/",a,i4.4,i2.2,i2.2,a)') trim(indir), trim(fname(1:index(trim(fname), '<')-1)), idate(1:3), &
      trim(fname(index(trim(fname), '>')+1:len(fname)))

    write (gol,'("[",a,"] ","Reading ",a)') trim(rname), trim(infile); call goPr

    inquire(file=infile, exist=exists_infile)

    if (exists_infile) then

      ! open file
      status = NF90_Open(infile, NF90_NOWRITE, ncid)
      IF_NF90_NOTOK_RETURN(status=1)
 
      ! get dimensions
      status = NF90_Inq_DimID(ncid, "sounding_id", dimid)
      IF_NF90_NOTOK_RETURN(status=1)

      status = NF90_Inquire_Dimension(ncid, dimid, len = nsoundings)
      IF_NF90_NOTOK_RETURN(status=1)

      ! determine type of sounding_id
      status = NF90_Inq_varID(ncid, "sounding_id", varid=varid)
      IF_NF90_NOTOK_RETURN(status=1)

      status = NF90_Inquire_variable(ncid, varid, xtype=sid_xtype)
      IF_NF90_NOTOK_RETURN(status=1)
      
      status = NF90_Close( ncid )
      IF_NF90_NOTOK_RETURN(status=1)
      
      call user_output_sounding_reset(status)
      IF_NOTOK_RETURN(status=1)

      ! variable allocations
      allocate(soundings%latitude(nsoundings))
      allocate(soundings%longitude(nsoundings))
      allocate(soundings%itau_sample(nsoundings))
      allocate(soundings%sec_1970(nsoundings))
      allocate(soundings%idate(6, nsoundings))
      allocate(soundings%mask(nsoundings))
      allocate(soundings%sounding_num(nsoundings))
      allocate(soundings%sounding_id_i8(nsoundings))
      allocate(soundings%sounding_id_r8(nsoundings))
      allocate(soundings%oro(nsoundings))
      allocate(soundings%region(nsoundings))
      allocate(soundings%blh(nsoundings))
      allocate(soundings%mf(ntracetloc, lm(region), nsoundings))
      allocate(soundings%temper(lm(region), nsoundings))
      allocate(soundings%phlb(lm(region)+1, nsoundings))
      allocate(soundings%m(lm(region), nsoundings))
      allocate(soundings%gph(lm(region)+1, nsoundings))
      allocate(soundings%wu(lm(region), nsoundings))
      allocate(soundings%wv(lm(region), nsoundings))

      ! variable initializations
      soundings%date = idate
      soundings%tau = itau
      soundings%mask = .false.
      soundings%sounding_num = -1

      ! get variables from infile
      call ReadFromFile(infile,MDF_NETCDF4,"latitude",soundings%latitude,.true.,status)
      IF_NOTOK_RETURN(status=1)

      call ReadFromFile(infile,MDF_NETCDF4,"longitude",soundings%longitude,.true.,status)
      IF_NOTOK_RETURN(status=1)

      call ReadFromFile(infile,MDF_NETCDF4,"time",soundings%sec_1970,.true.,status)
      IF_NOTOK_RETURN(status=1)

      if(sid_xtype==NF90_DOUBLE) then
         call ReadFromFile(infile,MDF_NETCDF4,"sounding_id",soundings%sounding_id_r8,.true.,status)
         IF_NOTOK_RETURN(status=1)
      else
         call ReadFromFile(infile,MDF_NETCDF4,"sounding_id",soundings%sounding_id_i8,.true.,status)
         IF_NOTOK_RETURN(status=1)
      endif

    else

      nsoundings = 0

    endif
          
    write (gol,'("[",a,"] ","Number of soundings in input file: ",i)') trim(rname), nsoundings; call goPr

    call date2tau([1970, 1, 1, 0, 0, 0], itau_1970)
    do isound = 1, nsoundings
       soundings%itau_sample(isound) = int(soundings%sec_1970(isound) + float(itau_1970))
      call tau2date(soundings%itau_sample(isound), soundings%idate(1:6,isound)) ! calendar dates calculated for checks only
      if(sid_xtype==NF90_DOUBLE) then
         write (gol,'("[",a,"] ","Debug: Sounding ID - ",f16.0," at ",f7.1,"N, ",f8.1,"E at ",i4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2)') &
              trim(rname), soundings%sounding_id_r8(isound),soundings%latitude(isound),soundings%longitude(isound), &
              soundings%idate(1:6,isound); call goPr   ! soundings%itau_sample(isound), 
      else
         write (gol,'("[",a,"] ","Debug: Sounding ID - ",i16," at ",f7.1,"N, ",f8.1,"E at ",i4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2)') &
              trim(rname), soundings%sounding_id_i8(isound),soundings%latitude(isound),soundings%longitude(isound), &
              soundings%idate(1:6,isound); call goPr   ! soundings%itau_sample(isound), 
      endif
    enddo ! nsoundings

    if (debug) then 
      do isound = 1, nsoundings
         if(sid_xtype .eq. NF90_DOUBLE) then
            write (gol,'("[",a,"] ","Debug: Sounding ID - ",f16.0," at ",i4.4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2", Sounding Num: ",i)') &
                 trim(rname), soundings%sounding_id_r8(isound),soundings%idate(1:6,isound), isound; call goPr
         else
            write (gol,'("[",a,"] ","Debug: Sounding ID - ",i16," at ",i4.4,"/",i2.2,"/",i2.2," ",i2.2,":",i2.2,":",i2.2", Sounding Num: ",i)') &
                 trim(rname), soundings%sounding_id_i8(isound),soundings%idate(1:6,isound), isound; call goPr
         endif
      enddo ! nsoundings
    endif

    ! ok
    status = 0

  end subroutine user_output_sounding_read

  subroutine user_output_sounding_sample(region, status)

    ! --------------------------------------------
    ! Sample/interpolate model fields
    ! --------------------------------------------

    use dims,             only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend, zbeg, zend, xcyc
    use dims,             only : ndyn, region_name, idate, itau, at, bt
    use datetime,         only : date2tau, tau2date
    use meteo,            only : phlb_dat, gph_dat, oro_dat, temper_dat, pu_dat, pv_dat, m_dat, humid_dat
    use global_data,      only : mass_dat, conv_dat
    use chem_param,       only : fscale, uscale
    use binas,            only : ae, twopi

    implicit none
  
    ! --- variables ------------------------------

    integer, intent(in)   :: region ! region number being sampled
    integer, intent(out)  :: status
    integer               :: isound ! soundings index 
    integer               :: itau_start, itau_end
    integer, dimension(6) :: idate_start, idate_end
    integer               :: i, j, l ! grid cell indices
    integer               :: n ! soundings level index
    integer               :: t ! tracer index 
    integer               :: is, js ! i, j, l index of grid cell in which column is located
    integer               :: isn, jsn ! i, j index of grid cell which is taken as neighbour for linear interpolation
    real                  :: ris, rjs ! deviation of column from center of the grid cell (-0.5 ... +0.5)
    real                  :: wcx, wcy ! x,y-weight of grid cell (1.0 ... 0.5) 
    real                  :: dxr, dyr ! x,y resolution (in degrees) for current region 

    real, dimension(im(region)) :: lon
    real, dimension(jm(region)) :: lat
    
    real, dimension(:,:,:), pointer   :: phlb ! pressure grid boundaries (Pa)
    real, dimension(:,:,:), pointer   :: gph ! geopotential height
    real, dimension(:,:,:,:), pointer :: rm ! tracer mass
    real, dimension(:,:,:,:), pointer :: rxm, rym, rzm ! slopes of tracer mass
    real, dimension(:,:,:), pointer   :: m ! air mass
    real, dimension(:,:), pointer     :: oro ! orography
    real, dimension(:,:,:), pointer   :: temper ! temperature
    real, dimension(:,:,:), pointer   :: pu ! mass flux x-direction (kg/s)
    real, dimension(:,:,:), pointer   :: pv ! mass flux y-direction (kg/s)
    real, dimension(:,:), pointer     :: blh ! boundary layer height (m)
    real, dimension(:,:,:), pointer   :: q ! water vapor mass mixing ratio (kg/kg)

    real    :: lat_j ! latitude of grid cell center js (degrees)
    real    :: lat_jn ! latitude of grid cell center jsn (degrees)
    real    :: ddx_j ! x-extension of grid cell js (m)
    real    :: ddx_jn ! x-extension of grid cell jsn (m)
    integer :: js_north, js_south ! j index of grid cell for neighbours for v-windspeed interpolation
    integer :: is_west ! i index of grid cell for neighbour for u-windspeed interpolation
    real    :: ddy ! y-extension of grid cell js (m)
 
    ! --- constants ------------------------------

    character(len=*), parameter :: rname = mname//'/user_output_sounding_sample'
    real            , parameter :: wcz = 1.0
    real            , parameter :: rls = 0.0

    ! --- begin ----------------------------------

    ! define samping window based on length of dynamic timestep
    itau_start = itau
    itau_end = itau + ndyn

    call tau2date(itau_start, idate_start)
    call tau2date(itau_end, idate_end)

    if(debug) then
       write (gol,'("[",a,"] ","Region: ",a,", Sampling Window: ",i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," - ",&
            i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2)') trim(rname), region_name(region), idate_start, idate_end; call goPr
    endif

    ! tracer 
    m => m_dat(region)%data  ! [imr+4,jmr+4,lmr] ; halo=2
    rm => mass_dat(region)%rm_t ! [imr+4,jmr+4,lmr,ntracetloc] ; halo=2
    rxm => mass_dat(region)%rxm_t ! [imr+4,jmr+4,lmr,ntracetloc] ; halo=2
    rym => mass_dat(region)%rym_t ! [imr+4,jmr+4,lmr,ntracetloc] ; halo=2
    rzm => mass_dat(region)%rzm_t ! [imr+4,jmr+4,lmr,ntracetloc] ; halo=2

    ! meteo
    phlb => phlb_dat(region)%data ! [imr+4,jmr+4,lmr+1] ; halo=2, boundary vals
    gph => gph_dat(region)%data ! [imr,jmr,lmr+1] ; boundary vals
    oro => oro_dat(region)%data(:,:,1) ! [imr,jmr]
    temper => temper_dat(region)%data ! [imr,jmr,lmr]
    pu => pu_dat(region)%data ! [imr+2,jmr+2,lmr+1] ; halo=1, boundary vals
    pv => pv_dat(region)%data ! [imr+2,jmr+2,lmr+1] ; halo=1, boundary vals
    blh => conv_dat(region)%blh ! [imr,jmr]
    q => humid_dat(region)%data ! [imr,jmr,lmr]

    dyr = dy/yref(region)
    dxr = dx/xref(region)

    lon = (/(xbeg(region) + (i - 1) * dxr + dxr * 0.5, i = 1, im(region))/)
    lat = (/(ybeg(region) + (j - 1) * dyr + dyr * 0.5, j = 1, jm(region))/)

    do isound = 1, nsoundings

      soundings%sounding_num(isound) = isound

      if ((soundings%longitude(isound) .gt. float(xbeg(region))) .and. (soundings%longitude(isound) .le. float(xend(region))) .and. &
          (soundings%latitude(isound) .gt. float(ybeg(region))) .and. (soundings%latitude(isound) .le. float(yend(region))) .and. & 
          (soundings%itau_sample(isound) .ge. itau_start) .and. (soundings%itau_sample(isound) .lt. itau_end) .and. .not. soundings%mask(isound)) then

         if(debug) then
            if(sid_xtype .eq. NF90_DOUBLE) then
               write (gol,'("[",a,"] ","Sampling: ",a,", Sounding ID - ",f16.0)') &
                    trim(rname), trim(region_name(region)), soundings%sounding_id_r8(isound); call goPr
            else
               write (gol,'("[",a,"] ","Sampling: ",a,", Sounding ID - ",i16.16)') &
                    trim(rname), trim(region_name(region)), soundings%sounding_id_i8(isound); call goPr
            endif
         endif

        soundings%mask(isound) = .true. 
        soundings%region(isound) = trim(region_name(region))

        ! --------------------------------------------
        ! compute indices
        ! --------------------------------------------

        ! avoid edges
        ris = (soundings%longitude(isound) - float(xbeg(region))) / dxr + 0.99999
        rjs = (soundings%latitude(isound) - float(ybeg(region))) / dyr + 0.99999
        is  = int(ris) ! i-index of grid cell in which column is located
        js  = int(rjs) ! j-index of grid cell in which column is located
        ris = ris - is - 0.5 ! x-deviation from the center of the grid cell (-0.5...+0.5)
        rjs = rjs - js - 0.5 ! y-deviation from the center of the grid cell (-0.5...+0.5)

        ! the neighbor for x-interpolation
        if (ris .gt. 0) then 
          isn = is + 1     
        else
          isn = is - 1
        endif

        ! the neighbor for y-interpolation
        if (rjs .gt. 0) then 
          jsn = js + 1     
        else
          jsn = js - 1
        endif

        ! x- / y-weighting of grid cell in which column is located
        wcx = (1.0 - abs(ris)) ! 1.0 ... 0.5
        wcy = (1.0 - abs(rjs)) ! 1.0 ... 0.5

        ! if index of neighbour is exceeding range of region then  set neighbor = current cell (i.e. no interpolation)
        ! in case of cyclic x-boundaries take corresponding cyclic i index
        if (jsn == 0) jsn = 1
        if (jsn == jm(region) + 1) jsn = jm(region)
        if (xcyc(region) == 0) then ! non-cyclic boundaries
          if (isn == 0) isn = 1
          if (isn == im(region) + 1) isn = im(region)
        else ! cyclic x-boundaries
          if (isn == 0) isn = im(region)
          if (isn == im(region) + 1) isn = 1
        end if

        soundings%temper(:,isound) = wcx * wcy * temper(is,js,:) + (1.0 - wcx) * wcy * temper(isn,js,:) + &
          wcx * (1.0 - wcy) * temper(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * temper(isn,jsn,:)

        soundings%phlb(:,isound) = wcx * wcy * phlb(is,js,:) + (1.0 - wcx) * wcy * phlb(isn,js,:) + &
          wcx * (1.0 - wcy) * phlb(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * phlb(isn,jsn,:)

        soundings%gph(:,isound) = wcx * wcy * gph(is,js,:) + (1.0 - wcx) * wcy * gph(isn,js,:) + &
          wcx * (1.0 - wcy) * gph(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * gph(isn,jsn,:)

        soundings%m(:,isound) = wcx * wcy * m(is,js,:) + (1.0 - wcx) * wcy * m(isn,js,:) + &
          wcx * (1.0 - wcy) * m(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * m(isn,jsn,:)

        do t = 1, ntracetloc
          select case(interp_method)
            case(2) ! interpolation using slopes  
              ! rm-value is obtained from rm + slopes, slope = rxm = (rm*dX/dx *deltaX/2)
              soundings%mf(t,:,isound) = rm(is,js,:,t) + 2.0 * (ris * rxm(is,js,:,t) + rjs * rym(is,js,:,t) + rls * rzm(is,js,:,t))
              soundings%mf(t,:,isound) = max(0.0, soundings%mf(t,:,isound)) 
              soundings%mf(t,:,isound) = soundings%mf(t,:,isound) / m(is,js,:) * fscale(t) * uscale(t)
            case default ! no interpolation/3D linear interpolation (same as meteo)
              soundings%mf(t,:,isound) = (wcx * wcy * rm(is,js,:,t) / m(is,js,:) + & 
                (1.0 - wcx) * wcy * rm(is,js,:,t) / m(isn,js,:) + wcx * (1.0 - wcy) * rm(is,js,:,t) / m(is,jsn,:) + &
                (1.0 - wcx) * (1.0 - wcy) * rm(is,js,:,t) / m(isn,jsn,:)) * fscale(t) * uscale(t)
          end select
        enddo ! ntracetloc

        ! interpolated model orography (this is consistent with existing
        ! user_ouput, but better way is 3D interpolation of full matrix and save
        ! surface oro?) 
        soundings%oro(isound) = wcx * wcy * oro(is,js) + & 
          (1.0 - wcx) * wcy * oro(isn,js) + &
          wcx * (1.0 - wcy) * oro(is,jsn) + &
          (1.0 - wcx) * (1.0 - wcy) * oro(isn,jsn)

        ! interpolated model blh
        soundings%blh(isound) = wcx * wcy * blh(is,js) + & 
          (1.0 - wcx) * wcy * blh(isn,js) + &
          wcx * (1.0 - wcy) * blh(is,jsn) + &
          (1.0 - wcx) * (1.0 - wcy) * blh(isn,jsn)

        ! zonal wind velocity (u-wind)
        !   pu is eastward flux through east grid box boundary
        !   windspeed_u_interpol wind component from east     
        lat_j = ybeg(region) + (js - 0.5) * dyr ! latitude of grid cell center js  (degrees)
        lat_jn = ybeg(region) + (jsn - 0.5) * dyr ! latitude of grid cell center jsn (degrees)
        ddx_j = ae * twopi * dxr / 360.0 * cos(lat_j * twopi / 360.0) ! x-extension of grid cell js (m)
        ddx_jn = ae * twopi * dxr / 360.0 * cos(lat_jn * twopi / 360.0) ! x-extension of grid cell jsn (m)

        is_west = is - 1
        if ( xcyc(region) == 0 ) then
          if ( is_west == 0 ) is_west = 1 ! non-cyclic boundaries
        else
          if ( is_west == 0 ) is_west = im(region) ! cyclic x-boundaries
        endif

        soundings%wu(:,isound) = -(((0.5 - ris) * pu(is_west,js,:) / m(is_west,js,:) + &
          (0.5 + ris) * pu(is,js,:)/ m(is,js,:)) * ddx_j * wcy * wcz + &  
          ((0.5 - ris) * pu(is_west,jsn,:) / m(is_west,jsn,:) + (0.5 + ris) * pu(is,jsn,:) / m(is,jsn,:)) * &
          ddx_jn * (1.0 - wcy) * wcz + ((0.5 - ris) * pu(is_west,js,:) / m(is_west,js,:) + &
          (0.5 + ris) * pu(is,js,:) / m(is,js,:)) * ddx_j * wcy * (1.0 - wcz) + &  
          ((0.5 - ris) * pu(is_west,jsn,:) / m(is_west,js,:) + (0.5 + ris) * pu(is,jsn,:) / m(is,jsn,:)) * &
          ddx_jn * (1.0 - wcy) * (1.0 - wcz))

        ! meridional wind velocity (v-wind)
        !   pv northward flux through south grid box boundary 
        !   windspeed_v_interpol wind component from north    
        ddy = ae * twopi * dyr / 360.0 ! y-extension of grid cell (m)

        js_south = js - 1
        if (js_south == 0) js_south = 1

        js_north = js + 1
        if (js_north == (jm(region) + 1)) js_north = jm(region)

        soundings%wv(:,isound) = -(((0.5 - rjs) * pv(is,js,:) / m(is,js_south,:) + &
          (0.5 + rjs) * pv(is,js_north,:) / m(is,js,:)) * wcx * wcz + &  
          ((0.5 - rjs) * pv(isn,js,:) / m(isn,js_south,:) + (0.5 + rjs) * pv(isn,js_north,:) / m(isn,js,:)) * &
          (1.0 - wcx) * wcz + ((0.5 - rjs) * pv(is,js,:) / m(is,js_south,:) + (0.5 + rjs) * pv(is,js_north,:) / m(is,js,:)) * &
          wcx * (1.0 - wcz) + ((0.5 - rjs) * pv(isn,js,:) / m(isn,js_south,:) + &
          (0.5 + rjs) * pv(isn,js_north,:) / m(isn,js,:)) * (1.0 - wcx) * (1.0-wcz)) * ddy 

      endif

    enddo ! nsoundings

    ! remove pointers
    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(oro)
    nullify(temper)
    nullify(pu)
    nullify(pv)
    nullify(gph)
    nullify(blh)
    nullify(phlb)
    nullify(q)

    ! ok
    status = 0

  end subroutine user_output_sounding_sample

  subroutine user_output_sounding_step(region, status)

    use dims,        only : ndyn, region_name, itaur, itaui, itau
    use dims,        only : idate, idatei, idatee
    use datetime,    only : tau2date

    implicit none

    ! --- variables ------------------------------
    
    integer, intent(out)  :: status
    integer, intent(in)   :: region
    integer, dimension(6) :: idate_region
    logical               :: do_sample
    logical               :: do_write
 
    ! --- constants ------------------------------
    
    character(len=*), parameter ::  rname = mname//'/user_output_sounding_step'

    ! --- begin ----------------------------------

    do_sample = .false.
    do_write = .false.

    ! sample at the beginning of every dynamic timestep
    if (modulo(itaur(region)-itaui, ndyn) .eq. 0) do_sample = .true.

    ! sample when each region equals the model time (samples in each half of the operator sequence) 
    ! note: not used because sampling is done after a call to sources/sinks but before mixing
    !if (itaur(region) .eq. itau) do_sample = .true.

    ! write daily files and reset
    if ((itaur(region)-itaui .gt. 0) .and. (modulo(itaur(region)-itaui, 86400) .eq. 0) .and. (soundings%tau .ne. itau)) do_write = .true. 

    call tau2date(itaur(region), idate_region)

    if (debug) then
      write (gol,'("[",a,"] Debug: ",a,", Date - ",i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," Sample:",l," Write:",l)') &
        trim(rname), region_name(region), idate_region, do_sample, do_write; call goPr
    endif

    if (do_write) then
      call user_output_sounding_write(status)
      IF_NOTOK_RETURN(status=1)
  
      call user_output_sounding_read(status)
      IF_NOTOK_RETURN(status=1)

      do_sample=.true.
    endif

    if (do_sample) then
      call user_output_sounding_sample(region, status)
      IF_NOTOK_RETURN(status=1)
    endif

  end subroutine user_output_sounding_step

  subroutine user_output_sounding_write(status)

    ! --------------------------------------------
    ! Write sampled output to a file, deallocate arrays.
    ! --------------------------------------------

    use dims,            only : xbeg, ybeg, xend, yend, lm
    use global_data,     only : outdir
    use MDF,             only : MDF_Create, MDF_Close, MDF_EndDef, MDF_INDEPENDENT, MDF_Var_Par_Access
    use MDF,             only : MDF_NETCDF4, MDF_REPLACE, MDF_UNLIMITED, MDF_GLOBAL
    use MDF,             only : MDF_INT, MDF_INT64,MDF_SHORT, MDF_FLOAT, MDF_DOUBLE, MDF_CHAR
    use MDF,             only : MDF_Def_Dim, MDF_Def_Var, MDF_Put_Var, MDF_Put_Att

#ifdef MPI
    use mpi_const,       only : my_real,mpi_comm_world,ierr,mpi_integer,mpi_character
    use mpi_const,       only : MPI_SUCCESS, MPI_INFO_NULL, localComm
    use mpi_comm,        only : barrier
#endif

    implicit none

    ! --- variables ------------------------------

    integer, intent(out) :: status
    character(len=1024)  :: outfile = ''
    integer              :: t, n, isound, ipe
    integer              :: ncid
    integer              :: dimid_nsoundings, dimid_nboundaries, dimid_nchar, dimid_nlevels, dimid_ntrace
    integer              :: varid_id, varid_time, varid_lat, varid_lon
    integer              :: varid_blh, varid_oro
    integer              :: varid_wv, varid_m, varid_gph, varid_wu, varid_phlb, varid_temper, varid_region, varid_q
    integer              :: varid_trace
    character(len=1024)  :: comment
    integer              :: ilevel,nlevels

    real*8 , dimension(:,:,:), allocatable :: tracers
    integer, dimension(:)    , allocatable :: recvcounts
    integer, dimension(:)    , allocatable :: offsets

    character(len=256)    :: notes 
    character(len=1024)   :: disclaimer
    character(len=256)    :: email
    character(len=256)    :: url
    character(len=256)    :: institution
    character(len=256)    :: conventions
    character(len=256)    :: history
    character(len=256)    :: source 
    character(len=256)    :: sysdate 
    character(len=256)    :: progstring
    integer, dimension(8) :: isysdate

    ! --- constants ------------------------------

    integer         , parameter :: region = 1 ! reference for ncdf writes is the global region  
    character(len=*), parameter ::  rname = mname//'/user_output_sounding_write'

    nlevels=lm(region)

    ! --- begin ----------------------------------

    if (nsoundings == 0) return
    if (count(soundings%mask) == 0) return

    write (*,'("par_barrier up: myid = ",i)') myid
    call par_barrier
    write (*,'("par_barrier down: myid = ",i)') myid

if(.false.) then

    if (myid .eq. root) then
       allocate(tracers(nlevels,ntracet,nsoundings))
       allocate(recvcounts(npes))
       allocate(offsets(npes))
       offsets(1)=0
       do ipe = 1, npes
          recvcounts(ipe)=ntracet_ar(ipe-1)
          if (ipe .gt. 1) then
             offsets(ipe)=sum(recvcounts(1:(ipe-1)))
          endif
       enddo
    else

       ! ifort's "-check all" complains severely if the mix_all, offsets, and recvcounts buffers
       ! are unallocated, despite the MPI spec's insistence that these buffers are only used
       ! by the root process.

       allocate(tracers(1,1,1))
       allocate(recvcounts(1))
       allocate(offsets(1))

    endif

#ifdef MPI

    ! -- all PEs use soundings%mf as a send buffer
    do isound=1,nsoundings
       do ilevel=1,nlevels
          ! MPI_GATHERV is how root gets all the mix_local values from other
          ! PEs.  Note that they are in rank-order, i.e. sorted by the myid
          ! value, and they include the values from the root process also.
          ! 
          ! Note that we use MPI_GATHERV (v for variable data amount for each
          ! node) because PEs can have differing numbers of tracers.
          call MPI_GATHERV(&
               soundings%mf(:,ilevel,isound), ntracetloc, MY_REAL, &
               tracers(:,ilevel,isound), recvcounts, offsets, MY_REAL, & 
               root, MPI_COMM_WORLD, ierr)
          if (.not. ierr .eq. MPI_SUCCESS) then
             TRACEBACK
             status=1
             return
          endif
       enddo
    enddo
#endif
endif

    if (myid .eq. root) then

       call date_and_time(values = isysdate)
       write (sysdate, '(i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," UTC")') &
            isysdate(1), isysdate(2), isysdate(3), isysdate(5), isysdate(6), isysdate(7)

       ! Standard file information for CarbonTracker output.
       email = "carbontracker.team@noaa.gov"
       url = "http://carbontracker.noaa.gov"
       institution = "NOAA Earth System Research Laboratory"
       conventions = "CF-1.4"
       call getarg(0, progstring, status)
       write(history,'("Created ",a," by ",a,".")') trim(sysdate),trim(progstring)

       write (outfile,'(a,"/","sounding_",i4.4,i2.2,i2.2,".nc")') trim(outdir), soundings%date(1:3)

       ! open file 
#if defined MPI && defined with_netcdf4_par 
       ! overwrite existing files (clobber), provide MPI stuff:
       call MDF_Create( trim(outfile), MDF_NETCDF4, MDF_REPLACE, ncid, status, &
            mpi_comm=localComm, mpi_info=MPI_INFO_NULL )
       if (status/=0) then
          write (gol,'("from creating NetCDF4 file for writing in parallel;")'); call goErr
          write (gol,'("MDF module not compiled with netcdf4 support ?")'); call goErr
          TRACEBACK; status=1; return
       end if
#else
       ! overwrite existing files (clobber)
       call MDF_Create( trim(outfile), MDF_NETCDF4, MDF_REPLACE, ncid, status )
       IF_NOTOK_RETURN(status=1)
#endif

       ! global attributes 
       call MDF_Put_Att(ncid, MDF_GLOBAL, "email", values=trim(email), status=status)
       IF_ERROR_RETURN(status=1)

       call MDF_Put_Att(ncid, MDF_GLOBAL, "url", values=trim(url), status=status)
       IF_ERROR_RETURN(status=1)

       call MDF_Put_Att(ncid, MDF_GLOBAL, "institution", values=trim(institution), status=status)
       IF_ERROR_RETURN(status=1)

       call MDF_Put_Att(ncid, MDF_GLOBAL, "conventions", values=trim(conventions), status=status)
       IF_ERROR_RETURN(status=1)

       call MDF_Put_Att(ncid, MDF_GLOBAL, "history", values=trim(history), status=status)
       IF_ERROR_RETURN(status=1)

       ! define dimensions
       call MDF_Def_Dim(ncid, 'nsoundings', MDF_UNLIMITED, dimid_nsoundings, status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Dim(ncid, 'ntrace', ntracet, dimid_ntrace, status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Dim(ncid, 'nlevels', lm(region), dimid_nlevels, status) ! note: lm(region) uses nlevels from global grid
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Dim(ncid, 'nboundaries', lm(region)+1, dimid_nboundaries, status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Dim(ncid, "character_length", 16, dimid_nchar, status) ! must be GE character lengths defined in column list derived type 
       IF_NOTOK_RETURN(status=1)

       ! variables/attributes
       if(sid_xtype .eq. NF90_DOUBLE) then
          call MDF_Def_Var(ncid, 'sounding_id', MDF_DOUBLE, (/dimid_nsoundings/), varid_id, status)
          IF_NOTOK_RETURN(status=1)
       else
          call MDF_Def_Var(ncid, 'sounding_id', MDF_INT64, (/dimid_nsoundings/), varid_id, status)
          IF_NOTOK_RETURN(status=1)
       endif
       call MDF_Put_Att(ncid, varid_id, "units", values="unitless", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_id, "long_name", values="profile sounding ID", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'region_name', MDF_CHAR, (/dimid_nchar, dimid_nsoundings/), varid_region, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_region, "units", values="unitless", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'time', MDF_DOUBLE, (/dimid_nsoundings/), varid_time, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_time, "units", values="unitless", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_time, "long_name", values="seconds since 1970-01-01 00:00:00", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'latitude', MDF_FLOAT, (/dimid_nsoundings/), varid_lat, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_lat, "units", values="degrees_north", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_lat, "range", values=(/ybeg(region), yend(region)/), status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'longitude', MDF_FLOAT, (/dimid_nsoundings/), varid_lon, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_lon, "units", values="degrees_east", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_lon, "range", values=(/xbeg(region), xend(region)/), status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'oro', MDF_DOUBLE, (/dimid_nsoundings/), varid_oro, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_oro, "units", values="m^s/s^2", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_oro, "long_name", values="orography", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_oro, "standard_name", values="surface geopotential", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'blh', MDF_DOUBLE, (/dimid_nsoundings/), varid_blh, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_blh, "units", values="meters", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_blh, "long_name", values="blh", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_blh, "standard_name", values="boundary layer height", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, "tracers", MDF_DOUBLE, (/dimid_nlevels, dimid_ntrace, dimid_nsoundings/), varid_trace, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_trace, "units", values="micromol mol-1", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_trace, "_FillValue", values=-1e+34, status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_trace, "long_name", values="mole fraction of carbon dioxide in air", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'pressure', MDF_DOUBLE, (/dimid_nboundaries, dimid_nsoundings/), varid_phlb, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_phlb, "units", values="pascals", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_phlb, "long_name", values="pressure", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'temperature', MDF_DOUBLE, (/dimid_nlevels, dimid_nsoundings/), varid_temper, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_temper, "units", values="kelvin", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_temper, "long_name", values="temperature", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'air_mass', MDF_DOUBLE, (/dimid_nlevels, dimid_nsoundings/), varid_m, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_m, "units", values="kilograms", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_m, "long_name", values="air mass", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_m, "standard_name", values="interpolated mass of air", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'specific_humidity', MDF_DOUBLE, (/dimid_nlevels, dimid_nsoundings/), varid_q, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_q, "units", values="(kg water)/(kg air)", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_q, "long_name", values="specific_humidity", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_q, "standard_name", values="specific_humidity", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'gph', MDF_DOUBLE, (/dimid_nboundaries, dimid_nsoundings/), varid_gph, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_gph, "units", values="meters", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_gph, "long_name", values="gph", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_gph, "standard_name", values="geopotential height", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'u-wind', MDF_DOUBLE, (/dimid_nlevels, dimid_nsoundings/), varid_wu, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wu, "units", values="meters/second", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wu, "long_name", values="u-wind", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wu, "standard_name", values="zonal component of the wind velocity", status=status)
       IF_NOTOK_RETURN(status=1)

       call MDF_Def_Var(ncid, 'v-wind', MDF_DOUBLE, (/dimid_nlevels, dimid_nsoundings/), varid_wv, status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wv, "units", values="meters/second", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wv, "long_name", values="v-wind", status=status)
       IF_NOTOK_RETURN(status=1)
       call MDF_Put_Att(ncid, varid_wv, "standard_name", values="meridional component of the wind velocity", status=status)
       IF_NOTOK_RETURN(status=1)

       ! end definitions
       call MDF_EndDef(ncid, status)
       IF_NOTOK_RETURN(status=1)

       ! set independent data mode (not all processes are writing):
       call MDF_Var_Par_Access( ncid, varid_id  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_lat  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_lon  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_oro  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_region  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_blh  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_m  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_time  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_phlb  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_q  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_temper  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_gph  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_wu  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)
       call MDF_Var_Par_Access( ncid, varid_wv  , MDF_INDEPENDENT, status )
       IF_NOTOK_MDF(fid=ncid)

       ! Write variables

       if ( myid == root ) then

          ! root only writes data for variables that are the 
          ! same across all cores.

          if(sid_xtype .eq. NF90_DOUBLE) then
             call MDF_Put_Var(ncid, varid_id, soundings%sounding_id_r8, status)
             IF_NOTOK_RETURN(status=1)
          else
             call MDF_Put_Var(ncid, varid_id, soundings%sounding_id_i8, status)
             IF_NOTOK_RETURN(status=1)
          endif

          call MDF_Put_Var(ncid, varid_lat, soundings%latitude, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_lon, soundings%longitude, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_oro, soundings%oro, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_region, (/(trim(soundings%region(isound)), isound=1, size(soundings%region))/), status )
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_blh, soundings%blh, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_m, soundings%m, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_time, soundings%sec_1970, status)
          IF_NOTOK_RETURN(status=1)


          call MDF_Put_Var(ncid, varid_phlb, soundings%phlb, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_q, soundings%q, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_temper, soundings%temper, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_gph, soundings%gph, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_wu, soundings%wu, status)
          IF_NOTOK_RETURN(status=1)

          call MDF_Put_Var(ncid, varid_wv, soundings%wv, status)
          IF_NOTOK_RETURN(status=1)

       endif

#ifdef MPI
       call MDF_Put_Var(ncid, varid_trace, tracers, status)
#else
       call MDF_Put_Var(ncid, varid_trace, soundings%mf, status)
#endif
       IF_NOTOK_RETURN(status=1)

       ! close file
       call MDF_Close(ncid, status)
       IF_NOTOK_RETURN(status=1)

       write (gol,'("[",a,"] Wrote ",a)') trim(rname), trim(outfile); call goPr
    endif
    ! ok

    if(allocated(tracers)) deallocate(tracers)
    if(allocated(recvcounts)) deallocate(recvcounts)
    if(allocated(offsets)) deallocate(offsets)

1   status = 0

  end subroutine user_output_sounding_write

  subroutine user_output_sounding_done(status)

    ! --------------------------------------------
    ! Deallocate arrays.
    ! --------------------------------------------

    implicit none

    ! --- variables ------------------------------
    
    integer, intent(out) :: status

    ! --- constants ------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/user_output_sounding_done'

    ! --- begin ----------------------------------

    call user_output_sounding_write(status)
    IF_NOTOK_RETURN(status=1)
  
    call user_output_sounding_reset(status)
    IF_NOTOK_RETURN(status=1)

    ! ok
    status = 0

  end subroutine user_output_sounding_done

  subroutine user_output_sounding_reset(status)
 
    implicit none
 
    ! --- variables ------------------------------
    
    integer, intent(out) :: status

    ! --- constants ------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/user_output_sounding_reset'

    ! --- begin ----------------------------------

    write (gol,'("[",a,"] Deallocating arrays")') trim(rname); call goPr

!    if (allocated(soundings%psurf)) deallocate(soundings%psurf)
!    if (allocated(soundings%SigmaB)) deallocate(soundings%SigmaB)
    if (allocated(soundings%sounding_num)) deallocate(soundings%sounding_num)
    if (allocated(soundings%sounding_id_r8)) deallocate(soundings%sounding_id_r8)
    if (allocated(soundings%sounding_id_i8)) deallocate(soundings%sounding_id_i8)
    if (allocated(soundings%itau_sample)) deallocate(soundings%itau_sample)
    if (allocated(soundings%mask)) deallocate(soundings%mask)
!    if (allocated(soundings%pressure)) deallocate(soundings%pressure)
    if (allocated(soundings%latitude)) deallocate(soundings%latitude)
    if (allocated(soundings%longitude)) deallocate(soundings%longitude)
    if (allocated(soundings%sec_1970)) deallocate(soundings%sec_1970)
    if (allocated(soundings%idate)) deallocate(soundings%idate)
    if (allocated(soundings%oro)) deallocate(soundings%oro)
    if (allocated(soundings%region)) deallocate(soundings%region)
    if (allocated(soundings%blh)) deallocate(soundings%blh)
    if (allocated(soundings%mf)) deallocate(soundings%mf)
    if (allocated(soundings%temper)) deallocate(soundings%temper)
    if (allocated(soundings%phlb)) deallocate(soundings%phlb)
    if (allocated(soundings%m))  deallocate(soundings%m)
    if (allocated(soundings%gph)) deallocate(soundings%gph)
    if (allocated(soundings%wu)) deallocate(soundings%wu)
    if (allocated(soundings%wv)) deallocate(soundings%wv)
    if (allocated(soundings%q)) deallocate(soundings%q)

    ! ok
    status = 0

  end subroutine user_output_sounding_reset

end module user_output_sounding
