!#################################################################
!
! MODULE: user_output_tccon
!
! PUBLIC SUBROUTINES: 
!   user_output_tccon_init
!   user_output_tccon_step
!   user_output_tccon_done
!
! PRIVATE SUBROUTINES: 
!   user_output_tccon_sample
!   user_output_tccon_write
!   user_output_tccon_read
!   user_output_tccon_reset
!
! DESCRIPTION: 
!   Write column tracer fields on interpolated levels at discrete
!   sample times. Called from user_output.F90. 
!
! REVISION HISTORY: 
!
!     Notes:
!     1. Developed with NF90 library to read input files. MDF
!        not currently capable of handling netCDF4 groups or int64.
!     2. "user_output_tccon_step" samples after the first call to vertical
!        diffusion "v" in the first half of the operator splitting sequence.
!        All tccons with a timestamp within the entire window of both
!        splitting sequences are sampled. The sampling is done only after the 
!        first call to "v" because it follows the "xyz" advection sequence, with 
!        the idea that whatever chemistry has been applied will have seem some 
!        influence of transport. Calls to user_output_tccon_step after 
!        source/sink/chemistry (SCS) sequences are ignored. Note that sampling 
!        after vertical diffusion is defined as an rc-key ("output.after.step : v").
!
!        ----------
!        Operator sequence, Table 1 (Krol et al., 2005).
!        * denotes a sample of the model
!        ----------
!                 t               t+dT/2
!        region 1 XYZ             V*S
!        region 2    XYZV*SCSVZYX
!                 t+dT/2                t+dT
!        region 1 CSVZYX
!        region 2         CSVZYXXYZVS
!
!     Required rc file keys:
!       output.tccon : [T/F]
!       output.tccon.indir   : [/path/to/netcdf/files]
!       output.tccon.ncfile   : [prefix]"<YYYYMMDD>"[suffix]".nc"
!
!     Example:
!       output.tccon : T
!       output.tccon.indir : /scratch2/portfolios/BMC/co2/input/acos
!       output.tccon.ncfile : acos_b35_L2lite_<YYYYMMDD>_r02.nc
!
!### macro's #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
!#################################################################

module user_output_tccon

  use dims, only : nregions
  use GO, only : gol, goErr, goPr
  use ParTools, only : myid, root
  use chem_param, only : ntracet, names

  implicit none

  private

  ! --- public routines ------------------------

  public :: user_output_tccon_init
  public :: user_output_tccon_step
  public :: user_output_tccon_done

  ! --- constants ------------------------------

  character(len=*), parameter ::  mname = 'user_output_tccon'
  logical, parameter :: debug = .false.
  integer, parameter :: interp_method = 2 ! interpolation method output to file
                                          !   1: no interpolation ('grd')
                                          !   2: interpolation of tracers based on slopes ('slp')
                                          !   3: 3D linear interpolations ('int')
  
  ! --- variables ------------------------------

  character(len=1024) :: indir
  character(len=1024) :: fname 
  integer :: nobs
  integer :: nlevels

  type tccons_fields
    character(len=16), dimension(:), allocatable :: sitename
    integer, dimension(:), allocatable :: itau_sample 
    integer, dimension(6) :: date 
    integer :: tau
    real(8), dimension(:), allocatable :: sec_1970, day_1970 
    integer, dimension(:,:), allocatable :: time
    character(len=10), dimension(:), allocatable :: region
    real, dimension(:), allocatable :: psurf 
    real, dimension(:), allocatable :: SigmaB 
    real, dimension(:), allocatable :: latitude
    real, dimension(:), allocatable :: longitude
    integer, dimension(:), allocatable :: tccon_num
    logical, dimension(:), allocatable :: mask
    real, dimension(:,:), allocatable :: pressure 
    real, dimension(:), allocatable :: oro ! orography
    real, dimension(:), allocatable :: blh ! boundary layer height (m)
    real, dimension(:,:), allocatable :: temper
    real, dimension(:,:), allocatable :: phlb 
    real, dimension(:,:,:), allocatable :: mf
    real, dimension(:,:), allocatable :: m
    real, dimension(:,:), allocatable :: q
    real, dimension(:,:), allocatable :: gph ! geopotential height
    real, dimension(:,:), allocatable :: wu ! zonal component, wind velocity (x-direction)
    real, dimension(:,:), allocatable :: wv ! meridional component, wind velocity (y-direction)
  endtype tccons_fields

  type(tccons_fields) :: tccons 

  ! --- begin ----------------------------------

  contains

  subroutine user_output_tccon_init(rcF, status)

    ! --------------------------------------------
    ! Read the rc file
    ! --------------------------------------------

    use GO, only : TrcFile, ReadRc

    implicit none

    ! --- variables ------------------------------

    type(TrcFile), intent(in) :: rcF
    integer, intent(out) :: status
 
    ! --- constants ------------------------------
    
    character(len=*), parameter :: rname = mname//'/user_output_tccon_init'
    
    ! --- begin ----------------------------------

    if (myid /= root) return

    call ReadRc(rcF, 'output.tccon.indir', indir, status)
    IF_NOTOK_RETURN(status=1)

    call ReadRc(rcF, 'output.tccon.ncfile', fname, status)
    IF_NOTOK_RETURN(status=1)

    call user_output_tccon_read(status)
    IF_NOTOK_RETURN(status=1)

    ! ok
    status = 0

  end subroutine user_output_tccon_init

  subroutine user_output_tccon_read(status)

    ! --------------------------------------------
    ! Read the input file for latitude, longitude, altitude (pressure), and time. 
    ! --------------------------------------------

    use datetime, only : date2tau, tau2date
    use dims, only : idate, itau, lm
    use NetCDF, only : NF90_Open, NF90_Close, NF90_Inq_DimID, NF90_Inquire_Dimension
    use NetCDF, only : NF90_Inq_NCID, NF90_Inq_VarID, NF90_Get_Var 
    use NetCDF, only : NF90_NOWRITE

    implicit none

    ! --- variables ------------------------------

    integer, intent(out) :: status
    character(len=1024) :: infile = ''
    logical :: exists_infile
    integer :: ncid, dimid, varid, grpid
    integer :: isound
    integer :: itau_1970
    real*8 :: sec_sample
    integer, dimension(6) :: date_sample
    character(len=16), dimension(:), allocatable :: sitename
 
    ! --- constants ------------------------------
    
    integer, parameter :: region = 1 ! reference region
    character(len=*), parameter :: rname = mname//'/user_output_tccon_read'
    
    ! --- begin ----------------------------------

    if (myid /= root) return

    write (infile,'(a,"/",a,i4.4,i2.2,i2.2,a)') trim(indir), trim(fname(1:index(trim(fname), '<')-1)), idate(1:3), &
      trim(fname(index(trim(fname), '>')+1:len(fname)))

    write (gol,'("[",a,"] ","Reading ",a)') trim(rname), trim(infile); call goPr

    inquire(file=infile, exist=exists_infile)

    if (exists_infile) then

      ! open file
      status = NF90_Open(infile, NF90_NOWRITE, ncid)
      IF_NOTOK_RETURN(status=1)
 
      ! get dimensions
      status = NF90_Inq_DimID(ncid, "time", dimid)
      IF_NOTOK_RETURN(status=1)
      write (gol,'("[",a,"] ","Dimid time ",i)') trim(rname), dimid; call goPr

      status = NF90_Inquire_Dimension(ncid, dimid, len = nobs)
      IF_NOTOK_RETURN(status=1)
      write (gol,'("[",a,"] ","nobs ",i)') trim(rname), nobs; call goPr 

      call user_output_tccon_reset(status)
      IF_NOTOK_RETURN(status=1)

      ! variable allocations
      allocate(tccons%psurf(nobs))
      allocate(tccons%SigmaB(nlevels))
      allocate(tccons%latitude(nobs))
      allocate(tccons%longitude(nobs))
      allocate(tccons%itau_sample(nobs))
      allocate(tccons%sec_1970(nobs))
      allocate(tccons%day_1970(nobs))
      allocate(tccons%time(6, nobs))
      allocate(tccons%mask(nobs))
      allocate(tccons%tccon_num(nobs))
      allocate(tccons%sitename(nobs))
      allocate(tccons%pressure(nlevels, nobs))
      allocate(tccons%q(nlevels, nobs))
      allocate(tccons%oro(nobs))
      allocate(tccons%region(nobs))
      allocate(tccons%blh(nobs))
      allocate(tccons%mf(ntracet, lm(region), nobs))
      allocate(tccons%temper(lm(region), nobs))
      allocate(tccons%phlb(lm(region)+1, nobs))
      allocate(tccons%m(lm(region), nobs))
      allocate(tccons%gph(lm(region)+1, nobs))
      allocate(tccons%wu(lm(region), nobs))
      allocate(tccons%wv(lm(region), nobs))

      allocate(sitename(nobs))
!      allocate(dupcheck(nobs))

      ! variable initializations
      tccons%date = idate
      tccons%tau = itau
      tccons%mask = .false.
      tccons%tccon_num = -1

      ! get variables from "root" group
      status = NF90_Inq_VarID(ncid, "lat_deg", varid)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Get_Var(ncid, varid, tccons%latitude)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Inq_VarID(ncid, "long_deg", varid)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Get_Var(ncid, varid, tccons%longitude)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Inq_VarID(ncid, "time", varid)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Get_Var(ncid, varid, tccons%sec_1970)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Inq_VarID(ncid, "time", varid)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Get_Var(ncid, varid, tccons%day_1970)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Inq_VarID(ncid, "sitename", varid)
      IF_NOTOK_RETURN(status=1)

      status = NF90_Get_Var(ncid, varid, sitename)
      IF_NOTOK_RETURN(status=1)

!      do isound= 1, nobs
!        write (gol,'("[",a,"] ",a)') trim(rname), sitename(isound); call goPr
!      enddo

!      status = NF90_Inq_VarID(ncid, "sitename", varid)
!      IF_NOTOK_RETURN(status=1)
!
!      status = NF90_Get_Var(ncid, varid, tccons%sitename)
!      IF_NOTOK_RETURN(status=1)

      status = NF90_Close( ncid )
      IF_NOTOK_RETURN(status=1)

    else

      nobs = 0
      nlevels = 0

    endif
          
    write (gol,'("[",a,"] ","Number of tccons in input file: ",i)') trim(rname), nobs; call goPr

    call date2tau([1970, 1, 1, 0, 0, 0], itau_1970)
    do isound = 1, nobs
!      write (gol,'("[",a,"] ","before",f17.5)') trim(rname), tccons%sec_1970(isound); call goPr
      tccons%sec_1970(isound) = tccons%sec_1970(isound) * 86400
!      write (gol,'("[",a,"] ",i)') trim(rname), tccons%region(isound); call goPr 
      tccons%itau_sample(isound) = int(tccons%sec_1970(isound) + float(itau_1970))
      call tau2date(tccons%itau_sample(isound), tccons%time(1:6,isound)) ! calendar dates calculated for checks only
      write (tccons%sitename(isound), '(a)') sitename(isound) ! convert int64 array to string array
    enddo ! nobs

    if (debug) then 
      do isound = 1, nobs
        write (gol,'("[",a,"] ","Debug: Sounding ID - ",i4.4,i2.2,i2.2,i2.2,i2.2,i2.2,", Sounding Num: ",i)') &
          trim(rname), tccons%time(1:6,isound), isound; call goPr
      enddo ! nobs
    endif

    if(allocated(sitename)) then
       deallocate(sitename)
    endif

    ! ok
    status = 0

  end subroutine user_output_tccon_read

  subroutine user_output_tccon_sample(region, status)

    ! --------------------------------------------
    ! Sample/interpolate model fields
    ! --------------------------------------------

    use dims, only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend, zbeg, zend, xcyc
    use dims, only : ndyn, region_name, idate, itau, at, bt
    use datetime, only : date2tau, tau2date
    use meteo, only : phlb_dat, gph_dat, oro_dat, temper_dat, pu_dat, pv_dat, m_dat, humid_dat
    use global_data, only : mass_dat, conv_dat
    use chem_param, only : fscale, uscale
    use binas, only : ae, twopi

    implicit none
  
    ! --- variables ------------------------------

    integer, intent(in) :: region ! region number being sampled
    integer, intent(out) :: status
    integer :: isound ! tccons index 
    integer :: itau_start, itau_end
    integer, dimension(6) :: idate_start, idate_end
    integer :: i, j, l ! grid cell indices
    integer :: n ! tccons level index
    integer :: t ! tracer index 
    integer :: is, js ! i, j, l index of grid cell in which column is located
    integer :: isn, jsn ! i, j index of grid cell which is taken as neighbour for linear interpolation
    real :: ris, rjs ! deviation of column from center of the grid cell (-0.5 ... +0.5)
    real :: wcx, wcy ! x,y-weight of grid cell (1.0 ... 0.5) 
    real :: dxr, dyr ! x,y resolution (in degrees) for current region 

    real, dimension(im(region)) :: lon
    real, dimension(jm(region)) :: lat
    
    real, dimension(:,:,:), pointer :: phlb ! pressure grid boundaries (Pa)
    real, dimension(:,:,:), pointer :: gph ! geopotential height
    real, dimension(:,:,:,:), pointer :: rm ! tracer mass
    real, dimension(:,:,:,:), pointer :: rxm, rym, rzm ! slopes of tracer mass
    real, dimension(:,:,:), pointer :: m ! air mass
    real, dimension(:,:), pointer :: oro ! orography
    real, dimension(:,:,:), pointer :: temper ! temperature
    real, dimension(:,:,:), pointer :: pu ! mass flux x-direction (kg/s)
    real, dimension(:,:,:), pointer :: pv ! mass flux y-direction (kg/s)
    real, dimension(:,:), pointer :: blh ! boundary layer height (m)
    real, dimension(:,:,:), pointer :: q ! water vapor mass mixing ratio (kg/kg)

    real :: lat_j ! latitude of grid cell center js (degrees)
    real :: lat_jn ! latitude of grid cell center jsn (degrees)
    real :: ddx_j ! x-extension of grid cell js (m)
    real :: ddx_jn ! x-extension of grid cell jsn (m)
    integer :: js_north, js_south ! j index of grid cell for neighbours for v-windspeed interpolation
    integer :: is_west ! i index of grid cell for neighbour for u-windspeed interpolation
    real :: ddy ! y-extension of grid cell js (m)
 
    ! --- constants ------------------------------

    character(len=*), parameter :: rname = mname//'/user_output_tccon_sample'
    real, parameter :: wcz = 1.0
    real, parameter :: rls = 0.0

    ! --- begin ----------------------------------

    if (myid /= root) return

    ! define samping window based on length of dynamic timestep
    itau_start = itau
    itau_end = itau + ndyn

    call tau2date(itau_start, idate_start)
    call tau2date(itau_end, idate_end)

    if(debug) then
       write (gol,'("[",a,"] ","Region: ",a,", Sampling Window: ",i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," - ",&
            i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2)') trim(rname), region_name(region), idate_start, idate_end; call goPr
    endif

    ! tracer 
    m => m_dat(region)%data  ! [imr+4,jmr+4,lmr] ; halo=2
    rm => mass_dat(region)%rm_t ! [imr+4,jmr+4,lmr,ntracet] ; halo=2
    rxm => mass_dat(region)%rxm_t ! [imr+4,jmr+4,lmr,ntracet] ; halo=2
    rym => mass_dat(region)%rym_t ! [imr+4,jmr+4,lmr,ntracet] ; halo=2
    rzm => mass_dat(region)%rzm_t ! [imr+4,jmr+4,lmr,ntracet] ; halo=2

    ! meteo
    phlb => phlb_dat(region)%data ! [imr+4,jmr+4,lmr+1] ; halo=2, boundary vals
    gph => gph_dat(region)%data ! [imr,jmr,lmr+1] ; boundary vals
    oro => oro_dat(region)%data(:,:,1) ! [imr,jmr]
    temper => temper_dat(region)%data ! [imr,jmr,lmr]
    pu => pu_dat(region)%data ! [imr+2,jmr+2,lmr+1] ; halo=1, boundary vals
    pv => pv_dat(region)%data ! [imr+2,jmr+2,lmr+1] ; halo=1, boundary vals
    blh => conv_dat(region)%blh ! [imr,jmr]
    q => humid_dat(region)%data ! [imr,jmr,lmr]

    dyr = dy/yref(region)
    dxr = dx/xref(region)

    lon = (/(xbeg(region) + (i - 1) * dxr + dxr * 0.5, i = 1, im(region))/)
    lat = (/(ybeg(region) + (j - 1) * dyr + dyr * 0.5, j = 1, jm(region))/)

    do isound = 1, nobs

      tccons%tccon_num(isound) = isound
      tccons%pressure(:,isound) = tccons%psurf(isound) * tccons%SigmaB

      if ((tccons%longitude(isound) .gt. float(xbeg(region))) .and. (tccons%longitude(isound) .le. float(xend(region))) .and. &
          (tccons%latitude(isound) .gt. float(ybeg(region))) .and. (tccons%latitude(isound) .le. float(yend(region))) .and. & 
          (tccons%itau_sample(isound) .ge. itau_start) .and. (tccons%itau_sample(isound) .lt. itau_end) .and. .not. tccons%mask(isound)) then

         if(debug) then
            write (gol,'("[",a,"] ","Sampling: ",a,", Sounding ID - ",a)') &
                 trim(rname), trim(region_name(region)), tccons%sitename(isound); call goPr
         endif

        tccons%mask(isound) = .true. 
        tccons%region(isound) = trim(region_name(region))

        ! --------------------------------------------
        ! compute indices
        ! --------------------------------------------

        ! avoid edges
        ris = (tccons%longitude(isound) - float(xbeg(region))) / dxr + 0.99999
        rjs = (tccons%latitude(isound) - float(ybeg(region))) / dyr + 0.99999
        is  = int(ris) ! i-index of grid cell in which column is located
        js  = int(rjs) ! j-index of grid cell in which column is located
        ris = ris - is - 0.5 ! x-deviation from the center of the grid cell (-0.5...+0.5)
        rjs = rjs - js - 0.5 ! y-deviation from the center of the grid cell (-0.5...+0.5)

        ! the neighbor for x-interpolation
        if (ris .gt. 0) then 
          isn = is + 1     
        else
          isn = is - 1
        endif

        ! the neighbor for y-interpolation
        if (rjs .gt. 0) then 
          jsn = js + 1     
        else
          jsn = js - 1
        endif

        ! x- / y-weighting of grid cell in which column is located
        wcx = (1.0 - abs(ris)) ! 1.0 ... 0.5
        wcy = (1.0 - abs(rjs)) ! 1.0 ... 0.5

        ! if index of neighbour is exceeding range of region then  set neighbor = current cell (i.e. no interpolation)
        ! in case of cyclic x-boundaries take corresponding cyclic i index
        if (jsn == 0) jsn = 1
        if (jsn == jm(region) + 1) jsn = jm(region)
        if (xcyc(region) == 0) then ! non-cyclic boundaries
          if (isn == 0) isn = 1
          if (isn == im(region) + 1) isn = im(region)
        else ! cyclic x-boundaries
          if (isn == 0) isn = im(region)
          if (isn == im(region) + 1) isn = 1
        end if

        tccons%temper(:,isound) = wcx * wcy * temper(is,js,:) + (1.0 - wcx) * wcy * temper(isn,js,:) + &
          wcx * (1.0 - wcy) * temper(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * temper(isn,jsn,:)

        tccons%phlb(:,isound) = wcx * wcy * phlb(is,js,:) + (1.0 - wcx) * wcy * phlb(isn,js,:) + &
          wcx * (1.0 - wcy) * phlb(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * phlb(isn,jsn,:)

        tccons%gph(:,isound) = wcx * wcy * gph(is,js,:) + (1.0 - wcx) * wcy * gph(isn,js,:) + &
          wcx * (1.0 - wcy) * gph(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * gph(isn,jsn,:)

        tccons%m(:,isound) = wcx * wcy * m(is,js,:) + (1.0 - wcx) * wcy * m(isn,js,:) + &
          wcx * (1.0 - wcy) * m(is,jsn,:) + (1.0 - wcx) * (1.0 - wcy) * m(isn,jsn,:)

        do t = 1, ntracet
          select case(interp_method)
            case(2) ! interpolation using slopes  
              ! rm-value is obtained from rm + slopes, slope = rxm = (rm*dX/dx *deltaX/2)
              tccons%mf(t,:,isound) = rm(is,js,:,t) + 2.0 * (ris * rxm(is,js,:,t) + rjs * rym(is,js,:,t) + rls * rzm(is,js,:,t))
              tccons%mf(t,:,isound) = max(0.0, tccons%mf(t,:,isound)) 
              tccons%mf(t,:,isound) = tccons%mf(t,:,isound) / m(is,js,:) * fscale(t) * uscale(t)
            case default ! no interpolation/3D linear interpolation (same as meteo)
              tccons%mf(t,:,isound) = (wcx * wcy * rm(is,js,:,t) / m(is,js,:) + & 
                (1.0 - wcx) * wcy * rm(is,js,:,t) / m(isn,js,:) + wcx * (1.0 - wcy) * rm(is,js,:,t) / m(is,jsn,:) + &
                (1.0 - wcx) * (1.0 - wcy) * rm(is,js,:,t) / m(isn,jsn,:)) * fscale(t) * uscale(t)
          end select
        enddo ! ntracet

        ! interpolated model orography (this is consistent with existing
        ! user_ouput, but better way is 3D interpolation of full matrix and save
        ! surface oro?) 
        tccons%oro(isound) = wcx * wcy * oro(is,js) + & 
          (1.0 - wcx) * wcy * oro(isn,js) + &
          wcx * (1.0 - wcy) * oro(is,jsn) + &
          (1.0 - wcx) * (1.0 - wcy) * oro(isn,jsn)

        ! interpolated model blh
        tccons%blh(isound) = wcx * wcy * blh(is,js) + & 
          (1.0 - wcx) * wcy * blh(isn,js) + &
          wcx * (1.0 - wcy) * blh(is,jsn) + &
          (1.0 - wcx) * (1.0 - wcy) * blh(isn,jsn)

        ! zonal wind velocity (u-wind)
        !   pu is eastward flux through east grid box boundary
        !   windspeed_u_interpol wind component from east     
        lat_j = ybeg(region) + (js - 0.5) * dyr ! latitude of grid cell center js  (degrees)
        lat_jn = ybeg(region) + (jsn - 0.5) * dyr ! latitude of grid cell center jsn (degrees)
        ddx_j = ae * twopi * dxr / 360.0 * cos(lat_j * twopi / 360.0) ! x-extension of grid cell js (m)
        ddx_jn = ae * twopi * dxr / 360.0 * cos(lat_jn * twopi / 360.0) ! x-extension of grid cell jsn (m)

        is_west = is - 1
        if ( xcyc(region) == 0 ) then
          if ( is_west == 0 ) is_west = 1 ! non-cyclic boundaries
        else
          if ( is_west == 0 ) is_west = im(region) ! cyclic x-boundaries
        endif

        tccons%wu(:,isound) = -(((0.5 - ris) * pu(is_west,js,:) / m(is_west,js,:) + &
          (0.5 + ris) * pu(is,js,:)/ m(is,js,:)) * ddx_j * wcy * wcz + &  
          ((0.5 - ris) * pu(is_west,jsn,:) / m(is_west,jsn,:) + (0.5 + ris) * pu(is,jsn,:) / m(is,jsn,:)) * &
          ddx_jn * (1.0 - wcy) * wcz + ((0.5 - ris) * pu(is_west,js,:) / m(is_west,js,:) + &
          (0.5 + ris) * pu(is,js,:) / m(is,js,:)) * ddx_j * wcy * (1.0 - wcz) + &  
          ((0.5 - ris) * pu(is_west,jsn,:) / m(is_west,js,:) + (0.5 + ris) * pu(is,jsn,:) / m(is,jsn,:)) * &
          ddx_jn * (1.0 - wcy) * (1.0 - wcz))

        ! meridional wind velocity (v-wind)
        !   pv northward flux through south grid box boundary 
        !   windspeed_v_interpol wind component from north    
        ddy = ae * twopi * dyr / 360.0 ! y-extension of grid cell (m)

        js_south = js - 1
        if (js_south == 0) js_south = 1

        js_north = js + 1
        if (js_north == (jm(region) + 1)) js_north = jm(region)

        tccons%wv(:,isound) = -(((0.5 - rjs) * pv(is,js,:) / m(is,js_south,:) + &
          (0.5 + rjs) * pv(is,js_north,:) / m(is,js,:)) * wcx * wcz + &  
          ((0.5 - rjs) * pv(isn,js,:) / m(isn,js_south,:) + (0.5 + rjs) * pv(isn,js_north,:) / m(isn,js,:)) * &
          (1.0 - wcx) * wcz + ((0.5 - rjs) * pv(is,js,:) / m(is,js_south,:) + (0.5 + rjs) * pv(is,js_north,:) / m(is,js,:)) * &
          wcx * (1.0 - wcz) + ((0.5 - rjs) * pv(isn,js,:) / m(isn,js_south,:) + &
          (0.5 + rjs) * pv(isn,js_north,:) / m(isn,js,:)) * (1.0 - wcx) * (1.0-wcz)) * ddy 

      endif
    enddo ! nobs

    ! remove pointers
    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(oro)
    nullify(temper)
    nullify(pu)
    nullify(pv)
    nullify(gph)
    nullify(blh)
    nullify(phlb)
    nullify(q)

    ! ok
    status = 0

  end subroutine user_output_tccon_sample

  subroutine user_output_tccon_step(region, status)

    use dims, only : ndyn, region_name, itaur, itaui, itau
    use dims, only : idate, idatei, idatee
    use datetime, only : tau2date

    implicit none

    ! --- variables ------------------------------
    
    integer, intent(out) :: status
    integer, intent(in) :: region
    integer, dimension(6) :: idate_region
    logical :: do_sample
    logical :: do_write
 
    ! --- constants ------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/user_output_tccon_step'

    ! --- begin ----------------------------------

    if (myid /= root) return

    do_sample = .false.
    do_write = .false.

    ! sample at the beginning of every dynamic timestep
    if (modulo(itaur(region)-itaui, ndyn) .eq. 0) do_sample = .true.

    ! sample when each region equals the model time (samples in each half of the operator sequence) 
    ! note: not used because sampling is done after a call to sources/sinks but before mixing
    !if (itaur(region) .eq. itau) do_sample = .true.

    ! write daily files and reset
    if ((itaur(region)-itaui .gt. 0) .and. (modulo(itaur(region)-itaui, 86400) .eq. 0) .and. (tccons%tau .ne. itau)) do_write = .true. 

    call tau2date(itaur(region), idate_region)

    if (debug) then
      write (gol,'("[",a,"] Debug: ",a,", Date - ",i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," Sample:",l," Write:",l)') &
        trim(rname), region_name(region), idate_region, do_sample, do_write; call goPr
    endif

    if (do_write) then
      call user_output_tccon_write(status)
      IF_NOTOK_RETURN(status=1)
  
      call user_output_tccon_read(status)
      IF_NOTOK_RETURN(status=1)
      do_sample = .true.
      ! something here for counting that step was sampled?
    endif

    if (do_sample) then
      call user_output_tccon_sample(region, status)
      IF_NOTOK_RETURN(status=1)
    endif

  end subroutine user_output_tccon_step

  subroutine user_output_tccon_write(status)

    ! --------------------------------------------
    ! Write sampled output to a file, deallocate arrays.
    ! --------------------------------------------

    use dims, only : xbeg, ybeg, xend, yend, lm
    use global_data, only : outdir
    use MDF, only : MDF_Create, MDF_Close, MDF_EndDef
    use MDF, only : MDF_NETCDF, MDF_REPLACE, MDF_UNLIMITED, MDF_GLOBAL
    use MDF, only : MDF_INT, MDF_SHORT, MDF_FLOAT, MDF_DOUBLE, MDF_CHAR
    use MDF, only : MDF_Def_Dim, MDF_Def_Var, MDF_Put_Var, MDF_Put_Att

    implicit none

    ! --- variables ------------------------------
    
    integer, intent(out) :: status
    character(len=1024) :: outfile = ''
    integer :: t, n, isound
    integer :: ncid
    integer :: dimid_nobs, dimid_nlevels, dimid_nboundaries, dimid_nchar
    integer :: varid_id, varid_num, varid_time, varid_lat, varid_lon
    integer :: varid_blh, varid_oro
    integer :: varid_wv, varid_m, varid_gph, varid_wu, varid_phlb, varid_temper, varid_region, varid_q
    integer, dimension(ntracet) :: varid_mf
    character(len=1024) :: comment

    character(len=256) :: notes 
    character(len=1024) :: disclaimer
    character(len=256) :: email
    character(len=256) :: url
    character(len=256) :: institution
    character(len=256) :: conventions
    character(len=256) :: history
    character(len=256) :: source 
    character(len=256) :: sysdate 
    character(len=256) :: progstring
    integer, dimension(8) :: isysdate

    ! --- constants ------------------------------
   
    integer, parameter :: region = 1 ! reference for ncdf writes is the global region  
    character(len=*), parameter  ::  rname = mname//'/user_output_tccon_write'

    ! --- begin ----------------------------------

    if (myid /= root) return
    if (nobs == 0) return
    if (count(tccons%mask) == 0) return

    call date_and_time(values = isysdate)
    write (sysdate, '(i4.4,"-",i2.2,"-",i2.2," ",i2.2,":",i2.2,":",i2.2," UTC")') &
           isysdate(1), isysdate(2), isysdate(3), isysdate(5), isysdate(6), isysdate(7)

    ! Standard file information for CarbonTracker output.
    email = "carbontracker.team@noaa.gov"
    url = "http://carbontracker.noaa.gov"
    institution = "NOAA Earth System Research Laboratory"
    conventions = "CF-1.4"
    call getarg(0, progstring, status)
    write(history,'("Created ",a," by ",a,".")') trim(sysdate),trim(progstring)

    write (outfile,'(a,"/","tccon_",i4.4,i2.2,i2.2,".nc")') trim(outdir), tccons%date(1:3)

    ! open file 
    call MDF_Create(trim(outfile), MDF_NETCDF, MDF_REPLACE, ncid, status)
    IF_NOTOK_RETURN(status=1)

    ! global attributes 
    call MDF_Put_Att(ncid, MDF_GLOBAL, "email", values=trim(email), status=status)
    IF_ERROR_RETURN(status=1)

    call MDF_Put_Att(ncid, MDF_GLOBAL, "url", values=trim(url), status=status)
    IF_ERROR_RETURN(status=1)

    call MDF_Put_Att(ncid, MDF_GLOBAL, "institution", values=trim(institution), status=status)
    IF_ERROR_RETURN(status=1)

    call MDF_Put_Att(ncid, MDF_GLOBAL, "conventions", values=trim(conventions), status=status)
    IF_ERROR_RETURN(status=1)

    call MDF_Put_Att(ncid, MDF_GLOBAL, "history", values=trim(history), status=status)
    IF_ERROR_RETURN(status=1)

    ! define dimensions
    call MDF_Def_Dim(ncid, 'nobs', MDF_UNLIMITED, dimid_nobs, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Dim(ncid, 'nlevels', lm(region), dimid_nlevels, status) ! note: lm(region) uses nlevels from global grid
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Dim(ncid, 'nboundaries', lm(region)+1, dimid_nboundaries, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Dim(ncid, "character_length", 16, dimid_nchar, status) ! must be GE character lengths defined in column list derived type 
    IF_NOTOK_RETURN(status=1)

    ! variables/attributes
    call MDF_Def_Var(ncid, 'sitename', MDF_CHAR, (/dimid_nchar, dimid_nobs/), varid_id, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_id, "units", values="longName", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_id, "long_name", values="TCCON site long name", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_id, "comment", values="from scan start time in UTC", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'region_name', MDF_CHAR, (/dimid_nchar, dimid_nobs/), varid_region, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_region, "units", values="none", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'time', MDF_DOUBLE, (/dimid_nobs/), varid_time, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_time, "units", values="days", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_time, "description", values="days since 1970-01-01 00:00:00 UTC", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'latitude', MDF_FLOAT, (/dimid_nobs/), varid_lat, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_lat, "units", values="degrees_north", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_lat, "range", values=(/ybeg(region), yend(region)/), status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'longitude', MDF_FLOAT, (/dimid_nobs/), varid_lon, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_lon, "units", values="degrees_east", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_lon, "range", values=(/xbeg(region), xend(region)/), status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'oro', MDF_DOUBLE, (/dimid_nobs/), varid_oro, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_oro, "units", values="m^s/s^2", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_oro, "long_name", values="orography", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_oro, "standard_name", values="surface geopotential", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'blh', MDF_DOUBLE, (/dimid_nobs/), varid_blh, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_blh, "units", values="meters", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_blh, "long_name", values="blh", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_blh, "standard_name", values="boundary layer height", status=status)
    IF_NOTOK_RETURN(status=1)

    do t = 1, ntracet
      select case(t)
        case(1)
          comment = "background - component due to initial condition (at 2000-01-01)"
        case(2)
          comment = "component due to fossil fuel burning (since 2000-01-01)"
        case(3)
          comment = "component due to air-sea gas exchange (since 2000-01-01)"
        case(4)
          comment = "component due to terrestrial biosphere exchange (since 2000-01-01)"
        case(5)
          comment = "component due to direct fire emissions (since 2000-01-01)"
        case default
          comment = "ERROR - no valid tracer found"
      end select

      call MDF_Def_Var(ncid, trim(names(t)), MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_mf(t), status)
      IF_NOTOK_RETURN(status=1)
      call MDF_Put_Att(ncid, varid_mf(t), "units", values="micromol mol-1", status=status)
      IF_NOTOK_RETURN(status=1)
      call MDF_Put_Att(ncid, varid_mf(t), "_FillValue", values=-1e+34, status=status)
      IF_NOTOK_RETURN(status=1)
      call MDF_Put_Att(ncid, varid_mf(t), "long_name", values="mole fraction of carbon dioxide in air", status=status)
      IF_NOTOK_RETURN(status=1)
      call MDF_Put_Att(ncid, varid_mf(t), "comment", values=trim(comment), status=status)
      IF_NOTOK_RETURN(status=1)
    enddo ! ntracet

    call MDF_Def_Var(ncid, 'pressure', MDF_DOUBLE, (/dimid_nboundaries, dimid_nobs/), varid_phlb, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_phlb, "units", values="pascals", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_phlb, "long_name", values="pressure", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'temperature', MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_temper, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_temper, "units", values="kelvin", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_temper, "long_name", values="temperature", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'air_mass', MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_m, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_m, "units", values="kilograms", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_m, "long_name", values="air mass", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_m, "standard_name", values="interpolated mass of air", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'specific_humidity', MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_q, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_q, "units", values="(kg water)/(kg air)", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_q, "long_name", values="specific_humidity", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_q, "standard_name", values="specific_humidity", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'gph', MDF_DOUBLE, (/dimid_nboundaries, dimid_nobs/), varid_gph, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_gph, "units", values="meters", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_gph, "long_name", values="gph", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_gph, "standard_name", values="geopotential height", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'u-wind', MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_wu, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wu, "units", values="meters/second", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wu, "long_name", values="u-wind", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wu, "standard_name", values="zonal component of the wind velocity", status=status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Def_Var(ncid, 'v-wind', MDF_DOUBLE, (/dimid_nlevels, dimid_nobs/), varid_wv, status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wv, "units", values="meters/second", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wv, "long_name", values="v-wind", status=status)
    IF_NOTOK_RETURN(status=1)
    call MDF_Put_Att(ncid, varid_wv, "standard_name", values="meridional component of the wind velocity", status=status)
    IF_NOTOK_RETURN(status=1)

    ! end definitions
    call MDF_EndDef(ncid, status)
    IF_NOTOK_RETURN(status=1)

    ! subtract tracer background
    !do n = 1, lm(region) 
    !  do t = 1, ntracet
    !    if (t .gt. 1) then 
    !        tccons%mf(t,n,:) = tccons%mf(t,n,:) - tccons%mf(1,n,:)
    !    endif
    !  enddo
    !enddo

    ! Write variables
!    call MDF_Put_Var(ncid, varid_num, tccons%tccon_num, status)
!    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_lat, tccons%latitude, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_lon, tccons%longitude, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_oro, tccons%oro, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_region, (/(trim(tccons%region(isound)), isound=1, size(tccons%region))/), status )
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_blh, tccons%blh, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_m, tccons%m, status)
    IF_NOTOK_RETURN(status=1)
 
    call MDF_Put_Var(ncid, varid_id, (/(trim(tccons%sitename(isound)), isound=1, size(tccons%sitename))/), status )
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_time, tccons%day_1970, status)
    IF_NOTOK_RETURN(status=1)

    do t = 1, ntracet
      call MDF_Put_Var(ncid, varid_mf(t), tccons%mf(t,:,:), status)
      IF_NOTOK_RETURN(status=1)
    enddo

    call MDF_Put_Var(ncid, varid_phlb, tccons%phlb, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_q, tccons%q, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_temper, tccons%temper, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_gph, tccons%gph, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_wu, tccons%wu, status)
    IF_NOTOK_RETURN(status=1)

    call MDF_Put_Var(ncid, varid_wv, tccons%wv, status)
    IF_NOTOK_RETURN(status=1)

    ! close file
    call MDF_Close(ncid, status)
    IF_NOTOK_RETURN(status=1)

    write (gol,'("[",a,"] Wrote ",a)') trim(rname), trim(outfile); call goPr

    ! ok
    status = 0

  end subroutine user_output_tccon_write

  subroutine user_output_tccon_done(status)

    ! --------------------------------------------
    ! Deallocate arrays.
    ! --------------------------------------------

    implicit none

    ! --- variables ------------------------------
    
    integer, intent(out) :: status

    ! --- constants ------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/user_output_tccon_done'

    ! --- begin ----------------------------------

    call user_output_tccon_write(status)
    IF_NOTOK_RETURN(status=1)
  
    call user_output_tccon_reset(status)
    IF_NOTOK_RETURN(status=1)

    ! ok
    status = 0

  end subroutine user_output_tccon_done

  subroutine user_output_tccon_reset(status)
 
    implicit none
 
    ! --- variables ------------------------------
    
    integer, intent(out) :: status

    ! --- constants ------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/user_output_tccon_reset'

    ! --- begin ----------------------------------

    write (gol,'("[",a,"] Deallocating arrays")') trim(rname); call goPr

    if (allocated(tccons%psurf)) deallocate(tccons%psurf)
    if (allocated(tccons%SigmaB)) deallocate(tccons%SigmaB)
    if (allocated(tccons%tccon_num)) deallocate(tccons%tccon_num)
    if (allocated(tccons%sitename)) deallocate(tccons%sitename)
    if (allocated(tccons%itau_sample)) deallocate(tccons%itau_sample)
    if (allocated(tccons%mask)) deallocate(tccons%mask)
    if (allocated(tccons%pressure)) deallocate(tccons%pressure)
    if (allocated(tccons%latitude)) deallocate(tccons%latitude)
    if (allocated(tccons%longitude)) deallocate(tccons%longitude)
    if (allocated(tccons%sec_1970)) deallocate(tccons%sec_1970)
    if (allocated(tccons%day_1970)) deallocate(tccons%day_1970)
    if (allocated(tccons%time)) deallocate(tccons%time)
    if (allocated(tccons%oro)) deallocate(tccons%oro)
    if (allocated(tccons%region)) deallocate(tccons%region)
    if (allocated(tccons%blh)) deallocate(tccons%blh)
    if (allocated(tccons%mf)) deallocate(tccons%mf)
    if (allocated(tccons%temper)) deallocate(tccons%temper)
    if (allocated(tccons%phlb)) deallocate(tccons%phlb)
    if (allocated(tccons%m))  deallocate(tccons%m)
    if (allocated(tccons%gph)) deallocate(tccons%gph)
    if (allocated(tccons%wu)) deallocate(tccons%wu)
    if (allocated(tccons%wv)) deallocate(tccons%wv)
    if (allocated(tccons%q)) deallocate(tccons%q)


    ! ok
    status = 0

  end subroutine user_output_tccon_reset

end module user_output_tccon
